from .graphql import ArgumentValueKind


## API Configurations
API_V2_ENDPOINT = 'https://api.monday.com/v2'
API_V2_FILE_ENDPOINT = 'https://api.monday.com/v2/file'
CONNECTION_TIMEOUT = 10


## Operation methods
# Boards
BOARDS = 'boards'
CREATE_BOARD = 'create_board'
ARCHIVE_BOARD = 'archive_board'
ADD_SUBSCRIBERS_TO_BOARD = 'add_subscribers_to_board'
DELETE_SUBSCRIBERS_FROM_BOARD = 'delete_subscribers_from_board'
# Columns
CREATE_COLUMN = 'create_column'
CHANGE_COLUMN_VALUE = 'change_column_value'
CHANGE_MULTIPLE_COLUMN_VALUES = 'change_multiple_column_values'
# Groups
DUPLICATE_GROUP = 'duplicate_group'
CREATE_GROUP = 'create_group'
ARCHIVE_GROUP = 'archive_group'
DELETE_GROUP = 'delete_group'
# Items
ITEMS = 'items'
ITEMS_BY_COLUMN_VALUES = 'items_by_column_values'
ITEMS_BY_MULTIPLE_COLUMN_VALUES = 'items_by_multiple_column_values'
CREATE_ITEM = 'create_item'
CREATE_SUBITEM = 'create_subitem'
CLEAR_ITEM_UPDATES = 'clear_item_updates'
MOVE_ITEM_TO_GROUP = 'move_item_to_group'
ARCHIVE_ITEM = 'archive_item'
DELETE_ITEM = 'delete_item'
DUPLICATE_ITEM = 'duplicate_item'
# Updates
UPDATES = 'updates'
CREATE_UPDATE = 'create_update'
DELETE_UPDATE = 'delete_update'
# Notifications
CREATE_NOTIFICATION = 'create_notification'
# Tags
TAGS = 'tags'
CREATE_OR_GET_TAG = 'create_or_get_tag'
# Users 
USERS = 'users'
# Teams
TEAMS = 'teams'
# Me
ME = 'me'
# Account
ACCOUNT = 'account'
# Webhooks
CREATE_WEBHOOK = 'create_webhook'
DELETE_WEBHOOK = 'delete_webhook'
# Workspace
CREATE_WORKSPACE = 'create_workspace'
# Files
ASSETS = 'assets'
ADD_FILE_TO_UPDATE = 'add_file_to_update'
ADD_FILE_TO_COLUMN = 'add_file_to_column'


## Default Fields
DEFAULT_BOARD_QUERY_FIELDS = [
    'id', 
    'name', 
    'board_folder_id', 
    'board_kind', 
    'description', 
    'permissions',
    'pos',
    'state',
    'workspace_id',
    'updated_at'
]

DEFAULT_ACTIVITY_LOG_QUERY_FIELDS = [
    'account_id',
    'created_at',
    'data',
    'entity',
    'event',
    'id',
    'user_id'
]

DEFAULT_BOARD_VIEW_QUERY_FIELDS = [
    'id',
    'name',
    'settings_str',
    'type'
]

DEFAULT_COLUMN_QUERY_FIELDS = [
    'id',
    'title',
    'settings_str',
    'archived',
    'type',
    'width'
]

DEFAULT_ITEM_QUERY_FIELDS = [
    'id',
    'name',
    'created_at',
    'creator_id',
    'state'
]

DEFAULT_GROUP_QUERY_FIELDS = [
    'id', 
    'title', 
    'archived', 
    'color', 
    'deleted', 
    'position'
]

DEFAULT_COLUMN_VALUE_QUERY_FIELDS = [
    'id',
    'title',
    'text',
    'value',
    'additional_info'
]

DEFAULT_TAG_QUERY_FIELDS = [
    'id',
    'name',
    'color'
]

DEFAULT_NOTIFICATION_QUERY_FIELDS = [
    'id',
    'text'
]

DEFAULT_UPDATE_QUERY_FIELDS = [
    'id',
    'body',
    'created_at',
    'creator_id',
    'item_id',
    'text_body',
    'updated_at'
]

DEFAULT_REPLY_QUERY_FIELDS = [
    'id',
    'body',
    'created_at',
    'creator_id',
    'text_body',
    'updated_at'
]

DEFAULT_ASSET_QUERY_FIELDS = [
    'id',
    'name',
    'url'
]

DEFAULT_WEBHOOK_QUERY_FIELDS = [
    'id',
    'board_id'
]

DEFAULT_WORKSPACE_QUERY_FIELDS = [
    'id',
    'name',
    'kind',
    'description'
]

DEFAULT_USER_QUERY_FIELDS = [
    'id',
    'name',
    'url',
    'email',
    'enabled',
    'birthday',
    'country_code',
    'is_guest',
    'is_pending',
    'join_date',
    'location',
    'mobile_phone',
    'phone',
    'photo_original',
    'photo_thumb',
    'photo_tiny',
    'time_zone_identifier',
    'title',
    'utc_hours_diff'
]

DEFAULT_TEAM_QUERY_FIELDS = [
    'id',
    'name',
    'picture_url'
]

DEFAULT_ACCOUNT_QUERY_FIELDS = [
    'id',
    'name',
    'first_day_of_the_week',
    'logo',
    'show_timeline_weekends',
    'slug'
]

DEFAULT_PLAN_QUERY_FIELDS = [
    'max_users',
    'period',
    'tier',
    'version'
]


##

## Optional Parameters

BOARDS_OPTIONAL_PARAMS = {
    'limit': ArgumentValueKind.Int,
    'page': ArgumentValueKind.Int,
    'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
    'board_kind': ArgumentValueKind.Enum,
    'state': ArgumentValueKind.Enum,
    'newest_first': ArgumentValueKind.Bool,
    'activity_logs': {
        'limit': ArgumentValueKind.Int,
        'page': ArgumentValueKind.Int,
        'user_ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
        'column_ids': (ArgumentValueKind.List, ArgumentValueKind.String),
        'group_ids': (ArgumentValueKind.List, ArgumentValueKind.String),
        'item_ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
        'from': ArgumentValueKind.String,
        'to': ArgumentValueKind.String
    },
    'columns': {
        'ids': (ArgumentValueKind.List, ArgumentValueKind.String)
    },
    'groups': {
        'ids': (ArgumentValueKind.List, ArgumentValueKind.String),
        'items': {
            'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
            'limit': ArgumentValueKind.Int,
            'page': ArgumentValueKind.Int
        }
    },
    'items': {
        'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
        'limit': ArgumentValueKind.Int,
        'page': ArgumentValueKind.Int
    },
    'updates': {
        'limit': ArgumentValueKind.Int,
        'page': ArgumentValueKind.Int
    },
    'views': {
        'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
        'type': ArgumentValueKind.String
    }
}

CREATE_BOARD_OPTIONAL_PARAMS = {
    'workspace_id': ArgumentValueKind.Int,
    'template_id': ArgumentValueKind.Int
}

ARCHIVE_BOARD_OPTIONAL_PARAMS = {

}

ADD_SUBSCRIBERS_TO_BOARD_OPTIONAL_PARAMS = {
    'kind': ArgumentValueKind.Enum
}

DELETE_SUBSCRIBERS_FROM_BOARD_OPTIONAL_PARAMS = {
    
}

CREATE_COLUMN_OPTIONAL_PARAMS = {

}

CHANGE_COLUMN_VALUE_OPTIONAL_PARAMS = {

}

CHANGE_MULTIPLE_COLUMN_VALUES_OPTIONAL_PARAMS = {

}

DUPLICATE_GROUP_OPTIONAL_PARAMS = {
    'add_to_top': ArgumentValueKind.Bool,
    'group_title': ArgumentValueKind.String
}

CREATE_GROUP_OPTIONAL_PARAMS = {

}

ARCHIVE_GROUP_OPTIONAL_PARAMS = {

}

DELETE_GROUP_OPTIONAL_PARAMS = {

}

ITEMS_OPTIONAL_PARAMS = {
    'assets': {
        'column_ids': (ArgumentValueKind.List, ArgumentValueKind.String)
    },
    'limit': ArgumentValueKind.Int,
    'page': ArgumentValueKind.Int,
    'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
    'newest_first': ArgumentValueKind.Bool,
    'column_values': {
        'ids': (ArgumentValueKind.List, ArgumentValueKind.Default),
    },
    'updates': {
        'limit': ArgumentValueKind.Int,
        'page': ArgumentValueKind.Int
    }
}

ITEMS_BY_COLUMN_VALUES_OPTIONAL_PARAMS = {
    'limit': ArgumentValueKind.Int,
    'page': ArgumentValueKind.Int,
    'column_type': ArgumentValueKind.String,
    'state': ArgumentValueKind.Enum,
    'column_values': {
        'ids': (ArgumentValueKind.List, ArgumentValueKind.Default),
    },
    'updates': {
        'limit': ArgumentValueKind.Int,
        'page': ArgumentValueKind.Int
    }
}

ITEMS_BY_MULTIPLE_COLUMN_VALUES_OPTIONAL_PARAMS = {
    'limit': ArgumentValueKind.Int,
    'page': ArgumentValueKind.Int,
    'column_type': ArgumentValueKind.String,
    'state': ArgumentValueKind.Enum
}

CREATE_ITEM_OPTIONAL_PARAMS = {
    'group_id': ArgumentValueKind.String,
    'column_values': ArgumentValueKind.Json
}

CREATE_SUBITEM_OPTIONAL_PARAMS = {
    'column_values': ArgumentValueKind.Json
}

MOVE_ITEM_TO_GROUP_OPTIONAL_PARAMS = {
    
}

ARCHIVE_ITEM_OPTIONAL_PARAMS = {

}

DELETE_ITEM_OPTIONAL_PARAMS = {

}

DUPLICATE_ITEM_ARGUMENTS = {
    'board_id': ArgumentValueKind.Int,
    'with_updates': ArgumentValueKind.Bool,
    'item_id': ArgumentValueKind.Int
}

UPDATES_OPTIONAL_PARAMS = {
    'limit': ArgumentValueKind.Int,
    'page': ArgumentValueKind.Int
}

CREATE_UPDATE_OPTIONAL_PARAMS = {
    'parent_id': ArgumentValueKind.Int
}

CREATE_NOTIFICATION_OPTIONAL_PARAMS ={
    'payload': ArgumentValueKind.Json
}

TAGS_OPTIONAL_PARAMS = {
    'ids': (ArgumentValueKind.List, ArgumentValueKind.Int)
}

CREATE_OR_GET_TAG_OPTIONAL_PARAMS = {
    'board_id': ArgumentValueKind.Int
}

USERS_OPTIONAL_PARAMS = {
    'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
    'kind': ArgumentValueKind.Enum,
    'newest_first': ArgumentValueKind.String,
    'limit': ArgumentValueKind.Int
}

TEAMS_OPTIONAL_PARAMS = {
    'ids': (ArgumentValueKind.List, ArgumentValueKind.Int)
}

CREATE_WEBHOOK_OPTIONAL_PARAMS = {
    'config': ArgumentValueKind.Json
}

CREATE_WORKSPACE_OPTIONAL_PARAMS = {
    'description': ArgumentValueKind.String
}

ASSETS_OPTIONAL_PARAMS = {
    
}


import json
from enum import Enum 


class OperationType(Enum):
    """GraphQL query operation type."""
    QUERY = 1
    MUTATION = 2


class ArgumentValueKind(Enum):
    """GraphQL types for arguments and variables."""
    Default = 0
    String = 1
    Int = 2
    Bool = 3
    Enum = 4
    List = 5
    Json = 6
    File = 7


class GraphQLNode():
    """Base node in a GraphQL query.
    
        Properties

            name : `str`
                Query node entity name.
            arguments : `dict`
                Arguments associated with the query node entity.

        Methods

            format_body : `str`
                Formats the GraphQL node into a query string.
            format_arguments : `str`
                Formats arguments as a query string.
    """

    def __init__(self, field_name: str):

        self.name: str = field_name
        self.arguments: dict = {}


    def format_body(self):
        """Formats the GraphQL node into a query string."""
        pass


    def format_arguments(self, body: str):
        """Formats arguments as query string.

            Parameters

                body : `str`
                    The name of the GraphQL field being formatted with arguments.

            Returns

                query_string : `str`
                    GraphQL-formatted query string.
        """

        formatted_args = ', '.join(['{}:{}'.format(key, value) for key, value in self.arguments.items()])
        formatted_args = formatted_args.replace("'", '"')
        return '{} ({})'.format(body, formatted_args)


class GraphQLField(GraphQLNode):
    """A GraphQL field node.

        Methods

            add_fields : `void`
                Add fields to node.
            add_arguments : `void`
                Add arguments to node.
            get_field : `moncli.api_v2.graphql.GraphQLField`
                Get a child field of the given GraphQL field.
            format_body : `str`
                Format the GraphQL node into a query string.
            format_children : `str`
                Format child fields into query string.
    """

    def __init__(self, field_name: str, field_map: dict, *args, **kwargs):
        super(GraphQLField, self).__init__(field_name)
        self.__children: dict = {}
        self._field_map = field_map
        
        self.add_fields(*args)
        self.add_arguments(**kwargs)


    def add_fields(self, *args):
        """Add fields to node.

            Parameters

                args : `tuple`
                    The list of child fields to add.
        """

        for field in args:

            if not field or field == '':
                continue
            
            if type(field) is str:
                field_split = field.split('.')
                parent_field = field_split[0]
                child_fields = field_split[1:]
                
                if not child_fields:
                    field_group = [[]]
                # Check for list notation.
                elif child_fields[0].startswith('[') and child_fields[0].endswith(']'):
                    if len(child_fields) > 1:
                        raise GraphQLError('Field list syntax is only available for leaf-level fields.')
                    
                    list_fields = child_fields[0][1:-1]
                    if list_fields == '*':
                        field_group = [[f] for f in self._field_map[parent_field]]
                    else:
                        field_group = [[f.strip()] for f in list_fields.split(',')]
                else:
                    field_group = [child_fields]
                
                for group in field_group:
                    existing_field = self.get_field(parent_field)
                    # Add the new fields to the existing field
                    if existing_field:
                        existing_field.add_fields('.'.join(group))
                        continue
                    
                    new_field = GraphQLField(field_split[0], self._field_map)
                    new_field.add_fields('.'.join(group))
                    self.__children.__setitem__(new_field.name, new_field)

            elif type(field) is GraphQLField:
                self.__children.__setitem__(field.name, field)


    def add_arguments(self, **kwargs):
        """Add arguments to node.

            Parameters

                kwargs : `dict`
                    The field's argument pairs.
        """

        for key, value in kwargs.items():
            if isinstance(value, ArgumentValue):
                arg_value : ArgumentValue = value
                formatted_value = arg_value.format()
                if formatted_value:
                    self.arguments.__setitem__(key, formatted_value)
            else:
                field = self.get_field(key)
                if not field:
                    raise GraphQLError('Unable to add arguments to "{}" because it was not specified as an output field'.format(key))

                self.get_field(key).add_arguments(**value)


    def get_field(self, path: str):
        """Get a child field of the given GraphQL field.

            Parameters

                path : `str`
                    The string field path of the field.

            Returns

                field : `moncli.api_v2.graphql.GraphQLField`
                    The retrieved graphql field.
        """

        split_path = path.split('.')
        child_name = split_path[0]
        
        if not self.__children.__contains__(child_name):
            return None

        node: GraphQLField = self.__children.get(child_name)
        remaining_path = split_path[1:]

        if (len(remaining_path) == 0):
            return node

        return node.get_field('.'.join(remaining_path))


    def format_body(self):
        """Format the GraphQL node into a query string."""

        body: str = self.name

        if len(self.arguments) > 0:
            body = self.format_arguments(body)

        if len(self.__children) > 0:
            body = self.format_children(body)

        return body

    
    def format_children(self, body: str):
        """Format child fields into query string.
            
            Parameters

                body : `str`
                    The body of the base query to be amended.

            Returns

                query_string : `str`
                    The amended GraphQL query string.

        """

        formatted_children = ', '.join([child.format_body() for child in self.__children.values()])
        return '{} {{ {} }}'.format(body, formatted_children)


class GraphQLOperation(GraphQLField):
    """A GraphQL node for operations.
    
        Properties
        
            action_type : `str`
                The action type of the operation (query / mutation)
            query_name : `str`
                The name of the executed command.
            args : `tuple`
                A collection of added fields.
            kwargs : `dict`
                A collection of argument mappings 

        Methods
        
            add_query_variable : `void`
                Add a query variable to the query.
    """

    def __init__(self, action_type: OperationType, query_name: str, field_map: dict, *args, **kwargs):

        self.action_type = action_type.name.lower()
        super(GraphQLOperation, self).__init__(query_name, field_map, *args, **kwargs)
        self.query_variables: dict = {}


    def add_query_variable(self, key: str, value):
        """Add a query variable to the query.

            Parameters

                key : `str`
                    The query variable key.
                value : `object`
                    The query variable value.
        """

        self.query_variables.__setitem__(key, value)  


    def format_body(self):
        """Format the GraphQL node into a query string."""

        body: str = self.name

        if len(self.arguments) > 0:
            body = self.format_arguments(body)

        body = self.format_children(body)

        if len(self.query_variables) > 0:
            var_list = ['${}: {}'.format(key, value) for key, value in self.query_variables.items()]
            var_format = '({})'.format(', '.join(var_list))
            body = '{} {} {{ {} }}'.format(self.action_type, var_format, body)
        else:
            body = '{} {{ {} }}'.format(self.action_type, body)

        return body


class ArgumentValue():
    """Base GraphQL argument type and format."""
    def __init__(self, value):
        self.value = value

    def format(self):
        return self.value


class StringValue(ArgumentValue):
    """GraphQL string argument type and format."""
    def __init__(self, value):
        super(StringValue, self).__init__(str(value))

    def format(self):
        return '"{}"'.format(self.value)


class IntValue(ArgumentValue):
    """GraphQL integer argument type and format."""
    def __init__(self, value):
        super(IntValue, self).__init__(int(value))

    def format(self):
        return self.value


class BoolValue(ArgumentValue):
    """GraphQL boolean argument type and format."""
    def __init__(self, value):
        super(BoolValue, self).__init__(value)

    def format(self):
        return str(self.value).lower()


class ListValue(ArgumentValue):
    """GraphQL list argument type and format."""
    def __init__(self, value):
        super(ListValue, self).__init__(list(value))

    def format(self):
        return self.value

class EnumValue(ArgumentValue):
    """GraphQL enum argument type and format."""
    def __init__(self, value):
        super(EnumValue, self).__init__(value)

    def format(self):
        enum : Enum = self.value
        return enum.name


class JsonValue(ArgumentValue):
    """GraphQL json argument type and format."""
    def __init__(self, value):
        super(JsonValue, self).__init__(value)

    def format(self):
        return json.dumps(json.dumps(self.value))


class FileValue(ArgumentValue):
    """GraphQL file argument type and format."""
    def __init__(self, value):
        super(FileValue, self).__init__(value)

    def format(self):
        return str(self.value)


class GraphQLError(Exception):
    """GraphQL client error."""
    def __init__(self, message: str):
        self.message = message


def create_field(field_name: str, *args, **kwargs):
    """Create a GraphQL field.

        Parameters

            field_name : `str`
                The name of the parent field to create.
            args : `tuple`
                A collection of child fields to add.
            kwargs : `dict`
                A field argument value map for queried return fields.

        Returns

            field : `moncli.api_v2.graphql.GraphQLField`
                A graphql field node.
    """

    return GraphQLField(field_name, *args, **kwargs)


def create_value(value, value_type: ArgumentValueKind):
    """Transform value into ArgumentValue
    
        Parameters

            value : `object`
                Any data object value.
            value_type : `moncli.api_v2.graphql.ArgumentValueKind`
                The argument value conversion type.

        Results

            value = `moncli.api_v2.graphql.ArgumentValue`
                The GraphQL argument value.
    """

    if value_type == ArgumentValueKind.Default:
        return ArgumentValue(value)
    elif value_type == ArgumentValueKind.String:
        return StringValue(value)   
    elif value_type == ArgumentValueKind.Int:
        return IntValue(value)
    elif value_type == ArgumentValueKind.Bool:
        return BoolValue(value)
    elif value_type == ArgumentValueKind.Enum:
        return EnumValue(value)
    elif value_type == ArgumentValueKind.List:
        return ListValue(value)
    elif value_type == ArgumentValueKind.Json:
        return JsonValue(value)


from . import constants
from .requests import execute_get, execute_post, execute_put, execute_delete

def get_boards(api_key, per_page = 25, only_globals = False, order_by_latest = False):

    resource_url = constants.BOARDS

    params = {
        'per_page': per_page,
        'only_globals': only_globals,
        'order_by_latest': order_by_latest
    }

    return execute_get(api_key, resource_url, params)


def post_board(api_key, user_id, name, description, board_kind):

    resource_url = constants.BOARDS

    body = {
        'user_id': user_id,
        'name': name,
        'description': description,
        'board_kind': board_kind
    }

    return execute_post(api_key, resource_url, body)


def get_board_by_id(api_key, board_id):

    resource_url = constants.BOARD_BY_ID.format(board_id)

    return execute_get(api_key, resource_url)


def delete_board_by_id(api_key, board_id):

    resource_url = constants.BOARD_BY_ID.format(board_id)

    return execute_delete(api_key, resource_url)


def get_board_groups(api_key, board_id, show_archived = False, show_deleted = False):

    resource_url = constants.BOARD_GROUPS.format(board_id)

    params = {
        'show_archived': show_archived,
        'show_deleted': show_deleted
    }

    return execute_get(api_key, resource_url, params)


def put_board_group(api_key, board_id, group_id, title, color):

    resource_url = constants.BOARD_GROUPS.format(board_id)

    body = {
        'group_id': group_id,
        'title': title,
        'color': color
    }

    return execute_put(api_key, resource_url, body)


def post_board_group(api_key, board_id, title):

    resource_url = constants.BOARD_GROUPS.format(board_id)

    body = {
        'title': title
    }

    return execute_post(api_key, resource_url, body)


def post_move_board_group(api_key, board_id, group_id, user_id, dest_board_id):

    resource_url = constants.BOARD_MOVE_GROUP.format(board_id, group_id)

    body = {
        'user_id': user_id,
        'dest_board_id': dest_board_id
    }

    return execute_post(api_key, resource_url, body)


def delete_board_group(api_key, board_id, group_id):

    resource_url = constants.BOARD_GROUP_BY_ID.format(board_id, group_id)

    return execute_delete(api_key, resource_url)


def get_board_columns(api_key, board_id, all_columns = False):

    resource_url = constants.BOARD_COLUMNS.format(board_id)

    params = {
        'all_columns': all_columns
    }

    return execute_get(api_key, resource_url, params)


def post_board_column(api_key, board_id, title, column_type, labels = None):

    resource_url = constants.BOARD_COLUMNS.format(board_id)

    body = {
        'title': title,
        'type': column_type
    }

    if column_type == 'status' and labels != None:
        body['labesl'] = labels

    return execute_post(api_key, resource_url, body)


def put_board_column(api_key, board_id, column_id, title = None, status = None, labels = None):

    resource_url = constants.BOARD_COLUMN_BY_ID.format(board_id, column_id)

    body = {}

    if title != None:
        body['title'] = title

    if status != None:
        body['status'] = status

    if labels != None:
        body['labels'] = labels

    return execute_put(api_key, resource_url, body)\


def delete_board_column(api_key, board_id, column_id):

    resource_url = constants.BOARD_COLUMN_BY_ID.format(board_id, column_id)

    return execute_delete(api_key, resource_url)


def get_board_column_value(api_key, board_id, column_id, pulse_id, return_as_array = False):

    resource_url = constants.BOARD_COLUMN_VALUE.format(board_id, column_id)

    params = {
        'pulse_id': pulse_id,
        'return_as_array': return_as_array
    }

    return execute_get(api_key, resource_url, params)


def put_board_text_column(api_key, board_id, column_id, pulse_id, text):

    resource_url = constants.BOARD_TEXT_COLUMN.format(board_id, column_id)

    body = {
        'pulse_id': pulse_id,
        'text': text,
    }

    return execute_put(api_key, resource_url, body)


def put_board_person_column(api_key, board_id, column_id, pulse_id, user_id):

    resource_url = constants.BOARD_PERSON_COLUMN.format(board_id, column_id)

    body = {
        'pulse_id': pulse_id,
        'user_id': user_id,
    }

    return execute_put(api_key, resource_url, body)


def put_board_status_column(api_key, board_id, column_id, pulse_id, color_index, update_id = None):

    resource_url = constants.BOARD_STATUS_COLUMN.format(board_id, column_id)

    body = {
        'pulse_id': pulse_id,
        'color_index': color_index,
    }

    if update_id != None:
        body['update_id'] = update_id

    return execute_put(api_key, resource_url, body)


def put_board_date_column(api_key, board_id, column_id, pulse_id, date_str):

    resource_url = constants.BOARD_DATE_COLUMN.format(board_id, column_id)

    body = {
        'pulse_id': pulse_id,
        'date_str': str(date_str)
    }

    return execute_put(api_key, resource_url, body)


def put_board_numeric_column(api_key, board_id, column_id, pulse_id, value):

    resource_url = constants.BOARD_NUMERIC_COLUMN.format(board_id, column_id)

    body = {
        'pulse_id': pulse_id,
        'value': str(value)
    }

    return execute_put(api_key, resource_url, body)


def put_board_tags_column(api_key, board_id, column_id, pulse_id, tags):

    resource_url = constants.BOARD_TAGS_COLUMN.format(board_id, column_id)

    body = {
        'pulse_id': pulse_id,
        'tags': ', '.join(tags)
    }

    return execute_put(api_key, resource_url, body)


def put_board_timeline_column(api_key, board_id, column_id, pulse_id, from_time, to_time):

    resource_url = constants.BOARD_TIMELINE_COLUMN.format(board_id, column_id)

    body = {
        'pulse_id': pulse_id,
        'from': from_time,
        'to': to_time
    }

    return execute_put(api_key, resource_url, body)


def get_board_pulses(api_key, board_id, page = 1, per_page = 25, order_by = None):

    resource_url = constants.BOARD_PULSES.format(board_id)

    params = {
        'page': page,
        'per_page': per_page
    }

    if order_by != None:
        params['order_by'] = order_by

    return execute_get(api_key, resource_url, params)


def post_board_pulse(
        api_key, 
        board_id, 
        user_id, 
        group_id, 
        name, 
        photo_from_url = None, 
        update_text = None, 
        add_to_bottom = False):

    resource_url = constants.BOARD_PULSES.format(board_id)

    body = {
        'user_id': user_id,
        'group_id': group_id,
        'pulse[name]': name,
        'add_to_bottom': add_to_bottom
    }

    if photo_from_url != None:
        body['photo_from_url'] = photo_from_url

    if update_text != None:
        body['update[text]'] = update_text

    return execute_post(api_key, resource_url, body)


def post_board_pulse_duplicate(api_key, board_id, pulse_id, group_id, owner_id):

    resource_url = constants.BOARD_DUPLICATE_PULSE.format(board_id, pulse_id)

    body = {
        'group_id': group_id,
        'owner_id': owner_id
    }

    return execute_post(api_key, resource_url, body)


def post_board_pulses_to_new_board(api_key, board_id, user_id, group_id, pulse_ids, dest_board_id = None, force_move_to_board = False):

    resource_url = constants.BOARD_MOVE_GROUP.format(board_id)

    body = {
        'user_id': user_id,
        'group_id': group_id,
        'pulse_ids': pulse_ids
    }

    if (dest_board_id != None):
        body['dest_board_id'] = dest_board_id
        body['force_move_to_board'] = force_move_to_board

    return execute_post(api_key, resource_url, body)


def get_board_subscribers(api_key, board_id, page = 1, per_page = 25, offset = 0):

    resource_url = constants.BOARD_SUBSCRIBERS.format(board_id)

    params = {
        'page': page,
        'per_page': per_page,
        'offset': offset
    }

    return execute_get(api_key, resource_url, params)


def put_board_subscriber(api_key, board_id, user_id, as_admin = False):

    resource_url = constants.BOARD_SUBSCRIBERS.format(board_id)

    body = {
        'as_admin': as_admin
    }

    return execute_put(api_key, resource_url, body)


def delete_board_subscriber_by_id(api_key, board_id, user_id):

    resource_url = constants.BOARD_SUBSCRIBERS_BY_ID.format(board_id, user_id)

    execute_delete(api_key, resource_url)


from schematics.models import Model
from schematics import types

from .. import api, entities as en, models as m
from ..error import GroupError
from ..models import MondayModel

class _Group(Model):
    """Group base model"""

    id = types.StringType(required=True)
    title = types.StringType()
    archived = types.BooleanType()
    color = types.StringType()
    deleted = types.BooleanType()
    position = types.StringType()


class Group(_Group):
    """ A group of items in a board.

        Properties

            archived : `bool`
                Is the group archived or not.
            color : `str`
                The group's color.
            deleted : `bool`
                Is the group deleted or not.
            id : `str`
                The group's unique identifier.
            items : `list[moncli.entities.Item]`
                The items in the group.
            position : `str`
                The group's position in the board.
            title : `str`
                The group's title.

        Methods

            duplicate : `moncli.entities.Group`
                Duplicate this group.
            archive : `moncli.entities.Group`
                Archives this group.
            delete : `moncli.entities.Group`
                Delete this group.
            add_item : `moncli.entities.Item`
                Add item to this group.
            get_items : `list[moncli.entities.Item]`
                Get items from this group.
    """

    def __init__(self, **kwargs):
        self.__creds = kwargs.pop('creds', None)
        self.__board = kwargs.pop('__board', None)
        self.__items = kwargs.pop('__items', None)

        items = kwargs.pop('items', None)
        if items != None and not self.__items:
            self.__items = [en.Item(creds=self.__creds, **item) for item in items]

        super(Group, self).__init__(kwargs)
    
    def __repr__(self):
        o = self.to_primitive()
        if self.__items:
            o['items'] = [item.to_primitive() for item in self.__items]
        return str(o)

    @property
    def items(self):
        """The items in the group."""
        if not self.__items:
            self.__items = self.get_items()
        return self.__items


    def duplicate(self, add_to_top: bool = False, *args, **kwargs):
        """Duplicate this group.

            Parameters

                args : `tuple`
                    The list of group fields to return.

            Returns

                group : `moncli.entities.Group`
                    The duplicated group.

            Return Fields

                archived : `bool`
                    Is the group archived or not.
                color : `str`
                    The group's color.
                deleted : `bool`
                    Is the group deleted or not.
                id : `str`
                    The group's unique identifier.
                items : `list[moncli.entities.Item]`
                    The items in the group.
                position : `str`
                    The group's position in the board.
                title : `str`
                    The group's title.
            
            Optional Arguments

                add_to_top : `bool`
                    Should the new group be added to the top.
                group_title : `str`
                    The group's title.
        """

        group_data = api.duplicate_group(
            self.__board.id, 
            self.id, 
            api_key=self.__creds.api_key_v2, 
            *args,
            **kwargs)
        return Group(
            creds=self.__creds,
            __board=self.__board.id,
            **group_data)


    def archive(self, *args):
        """Archives this group.

            Parameters

                args : `tuple`
                    The list of group fields to return.

            Returns

                group : `moncli.entities.Group`
                    The archived group.

            Return Fields

                archived : `bool`
                    Is the group archived or not.
                color : `str`
                    The group's color.
                deleted : `bool`
                    Is the group deleted or not.
                id : `str`
                    The group's unique identifier.
                items : `list[moncli.entities.Item]`
                    The items in the group.
                position : `str`
                    The group's position in the board.
                title : `str`
                    The group's title.
        """

        group_data = api.archive_group(
            self.__board.id,
            self.id, 
            *args,
            api_key=self.__creds.api_key_v2)
        return Group(
            creds=self.__creds,
            __board=self.__board,
            **group_data)


    def delete(self, *args):
        """Delete this group.

            Parameters

                args : `tuple`
                    The list of group fields to return.

            Returns

                item : `moncli.entities.Item`
                    The deleted item.

            Return Fields

                archived : `bool`
                    Is the group archived or not.
                color : `str`
                    The group's color.
                deleted : `bool`
                    Is the group deleted or not.
                id : `str`
                    The group's unique identifier.
                items : `list[moncli.entities.Item]`
                    The items in the group.
                position : `str`
                    The group's position in the board.
                title : `str`
                    The group's title.
        """

        group_data = api.delete_group(
            self.__board,
            self.id, 
            *args,
            api_key=self.__creds.api_key_v2)
        return Group(
            creds=self.__creds,
            __board=self.__board,
            **group_data)


    def add_item(self, item_name: str, as_model: type = None, *args, **kwargs):
        """Add item to this group.

            Parameters

                item_name : `str`
                    The new item's name.
                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    The list of item fields to return.
                kwargs : `dict`
                    Optional keyword arguments for adding an item to this group.

            Returns

                item : `moncli.entities.Item`
                    The newly added item to the group.

            Return Fields

                assets : `list[moncli.entities.set.Asset]`
                    The item's assets/files.
                board : `moncli.entities.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.Update`
                    The item's updates.
            
            Optional Arguments

                column_values : `json`
                    The column values of the new item.
        """

        item_data = api.create_item(
            item_name,
            self.__board.id, 
            *args,
            api_key=self.__creds.api_key_v2,
            group_id=self.id,
            **kwargs)
        items = en.Item(creds=self.__creds, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise GroupError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]

    def get_items(self, get_column_values: bool = True, as_model: type = None, *args, **kwargs):
        """Get items from this group.
    
            Parameters

                get_column_values: `bool`:
                    Retrieves item column values if set to `True`.
                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    The list of item fields to return.
        
            Returns

                items : `list[moncli.entities.Item]`
                    The collection of items belonging to the group.
        
            Return Fields

                assets : `list[moncli.entities.set.Asset]`
                    The item's assets/files.
                board : `moncli.entities.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.Update`
                    The item's updates.
            
            Optional Arguments

                limit : `int`
                    Number of items to get; the default is 25.
                page : `int`
                    Page number to get, starting at 1.
                ids : `list[str]`
                    A list of items unique identifiers.
                newest_first : `bool`
                    Get the recently created items at the top of the list.
        """

        if get_column_values:
            args = list(args)
            for arg in ['column_values.{}'.format(arg) for arg in api.DEFAULT_COLUMN_VALUE_QUERY_FIELDS]:
                if arg not in args:
                    args.append(arg)
            args.extend(['id', 'name'])
            
        group_kwargs = {'groups': {'ids': [self.id]}}
        if kwargs:
            group_kwargs['groups']['items'] = kwargs
        items_data = api.get_boards(
            *api.get_field_list(api.DEFAULT_ITEM_QUERY_FIELDS, 'groups.items', *args),
            api_key=self.__creds.api_key_v2, 
            ids=[int(self.__board.id)],
            limit=1,
            **group_kwargs)[0]['groups'][0]['items']
        items = [en.Item(creds=self.__creds, **item_data) for item_data in items_data] 
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise GroupError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]


api_key = None
connection_timeout = 10

from . import graphql as gql
from .exceptions import *
from .constants import *
from .handlers import *
from .requests import execute_query, upload_file, get_field_list, get_method_arguments


from .. import entities as en
from ..enums import *

## Column type mappings
COLUMN_TYPE_MAPPINGS = {
    'boolean': ColumnType.checkbox,
    'country': ColumnType.country,
    'date': ColumnType.date,
    'dropdown': ColumnType.dropdown,
    'dependency': ColumnType.dependency,
    'email': ColumnType.email,
    'hour': ColumnType.hour,
    'link': ColumnType.link,
    'location': ColumnType.location,
    'long-text': ColumnType.long_text,
    'name': ColumnType.name,
    'numeric': ColumnType.numbers,
    'multiple-person': ColumnType.people,
    'phone': ColumnType.phone,
    'pulse-updated': ColumnType.last_updated,
    'rating': ColumnType.rating,
    'color': ColumnType.status,
    'tag': ColumnType.tags,
    'team': ColumnType.team,
    'text': ColumnType.text,
    'timerange': ColumnType.timeline,
    'week': ColumnType.week,
    'timezone': ColumnType.world_clock,
    'file': ColumnType.file,
    'board-relation': ColumnType.board_relation,
    'subtasks': ColumnType.subitems,
    'pulse-log': ColumnType.creation_log
}


class BaseColumn(object):
    def __init__(self, **kwargs):
        self.id = kwargs.pop('id')
        self.title = kwargs.pop('title', None)
        self.settings_str = kwargs.pop('settings_str', None)


class Column(BaseColumn):
    """A board's column.
    
    Properties

        archived : `bool`
            Is the column archived or not.
        id : `str`
            The column's unique identifier.
        settings_str : `str`
            The column's settings in a string form.
        title : `str`
            The column's title.
        type : `str`
            The column's type.
        column_type : `moncli.entities.ColumnType`
            The column's type as an enum.
        width : `int`
            The column's width.
    """

    def __init__(self, **kwargs):
        self.archived = kwargs.pop('archived', None)
        self.type = kwargs.pop('type', None)
        self.width = kwargs.pop('width', None)
        super().__init__(**kwargs)
    
    def __repr__(self):
        return str(self.to_primitive())
    
    @property
    def column_type(self):
        # TODO - Find something else other than auto-number to default to.
        return COLUMN_TYPE_MAPPINGS.get(self.type, ColumnType.auto_number)

    def to_primitive(self):
        return dict(
            id=self.id,
            title=self.title,
            type=self.type,
            archived=self.archived,
            settings_str=self.settings_str,
            width=self.width)


class BaseColumnCollection(en.BaseCollection):
    
    def __init__(self, column_values: list = []):

        def get_index(index):
            if not isinstance(index, (int, str)):
                raise TypeError('Expected index type of int or str, got "{}" instead.'.format(index.__class__.__name__))
            if isinstance(index, (int,)):   
                return index
            for i in range(len(self._values)):
                column = self._values[i]
                if column.id == index or column.title == index:
                    return i
            raise KeyError('Collection contains no value for key "{}".'.format(index))

        super().__init__(column_values, BaseColumn, get_index)


class MoncliError(Exception):

    def __init__(self, error_code, entity_id, entity_type, message):
        super().__init__({
            'error_code': error_code,
            'entity_id': entity_id,
            'entity_type': entity_type,
            'message': message
        })


class MondayClientError(MoncliError):
    entity_type = 'MondayClient'

    def __init__(self, error_code, message):
        super().__init__(error_code, None, self.entity_type, message)
        
        
class BoardError(MoncliError):
    entity_type = 'Board'
    def __init__(self, error_code, entity_id, message):
        super().__init__(error_code, entity_id, self.entity_type, message)
        
        
class GroupError(MoncliError):
    entity_type = 'Group'
    def __init__(self, error_code, entity_id, message):
        super().__init__(error_code, entity_id, self.entity_type, message)


class ItemError(MoncliError):
    entity_type = 'Item'
    def __init__(self, error_code, entity_id, message):
        super().__init__(error_code, entity_id, self.entity_type, message)


class ColumnValueError(MoncliError):
    entity_type = 'ColumnValue'

    def __init__(self, error_code, entity_id, message):
        super().__init__(error_code, entity_id, self.entity_type, message)


import json, warnings
from moncli.entities.base import BaseCollection

from schematics.models import Model
from schematics.types import StringType, BooleanType, IntType, DictType, ListType, ModelType


class MondayClientCredentials():
    """monday.com client credentials.
    
    Properties

        api_key_v1 : `str`
            The access key for monday.com API v1.
        api_key_v2 : `str`
            The access key for monday.com API v2.
    """

    def __init__(self, api_key_v2: str = None):
        self.api_key_v2 = api_key_v2


class ActivityLog(Model):
    """monday.com client credentials.
    
    Properties

        account_id : `str`
            The user's payment account.
        created_at : `str`
            The activity log's created date.
        data : `str`
            The item's column values in string form.
        entity : `str`
            The monday.com object being updated.
        id : `str`
            The activity log's unique identifier.
        user_id : `str`
            The user who performed the change.
    """

    id = StringType(required=True)
    account_id = StringType()
    created_at = StringType()
    data = StringType()
    entity = StringType()
    event = StringType()
    user_id = StringType()

    def __repr__(self):
        return str(self.to_primitive())

class BoardView(Model):
    """A board's view.

    Properties

        id : `str`
            The view's unique identifier.
        name : `str`
            The view's name.
        settings_str : `str`
            The view's settings in a string from.
        type : `str`
            The view's type.
    """

    id = StringType(required=True)
    name = StringType()
    settings_str = StringType()
    type = StringType()

    def __repr__(self):
        return str(self.to_primitive())


class Notification(Model):
    """A notification.

    Properties

        id : `str`
            The notification`s unique identifier.
        text : `str`
            the notification text.
    """

    id = StringType(required=True)
    text = StringType()

    def __repr__(self):
        return str(self.to_primitive())


class Tag(Model):
    """A tag.

    Properties

        color : `str`
            The tag's color.
        id : `str`
            The tag's unique identifier.
        name : `str`
            The tag's name.
    """

    id = StringType(required=True)
    name = StringType()
    color = StringType()

    def __repr__(self):
        return str(self.to_primitive())


class Plan(Model):
    """ A payment plan.
    
    Properties

        max_users : `int`
            The maximum users allowed in the plan.
        period : `str`
            The plan's time period.
        tier : `str`
            The plan's tier.
        version : `int`
            The plan's version.
    """

    max_users = IntType()
    period = StringType()
    tier = StringType()
    version = IntType()

    def __repr__(self):
        return str(self.to_primitive())


class Webhook(Model):
    """ Monday webhooks

    Properties

        board_id : `str`
            The webhook's board id.
        id : `str`
            The webhook's unique identifier.
        is_active : `bool`
            If the webhook is still active.
    """

    id = StringType(required=True)
    board_id = StringType()
    is_active = BooleanType(default=False)

    def __repr__(self):
        return str(self.to_primitive())


class Workspace(Model):
    """ A monday.com workspace

    Properties

        description : `str`
            The workspace's description.
        id : `int`
            The workspace`s unique identifier.
        kind : `moncli.enums.WorkspaceKind`
            The workspace's kind (open / closed).
        name : `str`
            The workspace's name.
    """
    id = StringType(required=True)
    name = StringType()
    kind = StringType()
    description = StringType()

    def __repr__(self):
        return str(self.to_primitive())


from enum import Enum

class BoardKind(Enum):
    public = 1
    private = 2
    share = 3


class BoardsOrderBy(Enum):
    created_at = 1
    used_at = 2


class ColumnType(Enum):
    auto_number = 1
    checkbox = 2
    country = 3
    color_picker = 4
    creation_log = 5
    date = 6
    dropdown = 7
    email = 8
    hour = 9
    item_id = 10
    last_updated = 11
    link = 12
    location = 13
    long_text = 14
    numbers = 15
    people = 16
    phone = 17
    progress = 18
    rating = 19
    status = 20
    team = 21
    tags = 22
    text = 23
    timeline = 24
    time_tracking = 25
    vote = 26
    week = 27
    world_clock = 28
    file = 29
    subitems = 30
    board_relation = 31
    integration = 32
    name = 99999 # This column is a text type but comes standard with every board.
    dependency = 34

class NotificationTargetType(Enum):
    Project = 1
    Post = 2


# Board state enum
class State(Enum):
    all = 1
    active = 2
    archived = 3
    deleted = 4


class UserKind(Enum):
    all = 1
    non_guests = 2
    guests = 3
    non_pending = 4

    
class FirstDayOfTheWeek(Enum):
    sunday = 1
    monday = 2


class PeopleKind(Enum):
    person = 1
    team = 2


class WebhookEventType(Enum):
    create_item = 1
    change_column_value = 2
    change_specific_column_value = 3
    create_update = 4
    incoming_notification = 5
    change_name = 6
    create_subitem = 7
    change_subitem_column_value = 8
    change_subitem_name = 9
    create_subitem_update = 10


class WorkspaceKind(Enum):
    open = 1
    closed = 2


class SubscriberKind(Enum):
    subscriber = 1
    owner = 2


class WorkspaceSubscriberKind(Enum):
    subscriber = 0
    owner = 1


from .base import BaseCollection
from .objects import *
from .user import User, Team, Account
from .asset import Asset
from .update import Update, Reply
from .column import BaseColumn, Column, BaseColumnCollection
from . import column_value as cv
from .group import Group
from .board import Board, InvalidColumnValue
from .item import Item, UpdateNotFound
from .item import TooManyChangeSimpleColumnValueParameters,NotEnoughChangeSimpleColumnValueParameters
from .client import MondayClient
from .column_value.base import ColumnValue



from moncli.routes import constants
from .requests import execute_get, execute_post, execute_put, execute_delete

def get_pulses(
        api_key, 
        page = 1,
        per_page = 25, 
        offset = 0,
        order_by_latest = False,
        since = None,
        until = None):
    
    resource_url = constants.PULSES

    params = {
        'page': page,
        'per_page': per_page,
        'offset': offset,
        'order_by_latest': order_by_latest
    }

    if since != None:
        params['since'] = since

    if until != None:
        params['until'] = until

    return execute_get(api_key, resource_url, params)


def get_pulse_by_id(api_key, pulse_id):

    resource_url = constants.PULSE_BY_ID.format(pulse_id)

    return execute_get(api_key, resource_url)


def put_pulse_by_id(api_key, pulse_id, name):

    resource_url = constants.PULSE_BY_ID.format(pulse_id)

    body = {
        'name': name
    }

    return execute_put(api_key, resource_url, body)


def delete_pulse_by_id(api_key, pulse_id, archive = False):

    resource_url = constants.PULSE_BY_ID.format(pulse_id)

    params = {
        'archive': archive
    }

    return execute_delete(api_key, resource_url, params)


def get_pulse_subscribers(api_key, pulse_id, page = 1, per_page = 25, offset = 0):

    resource_url = constants.PULSE_SUBSCRIBERS.format(pulse_id)

    params = {
        'page': page,
        'per_page': per_page,
        'offset': offset
    }

    return execute_get(api_key, resource_url, params)


def put_pulse_subscriber(api_key, pulse_id, user_id, as_admin = False):

    resource_url = constants.PULSE_SUBSCRIBERS.format(pulse_id)

    body = {
        'user_id': user_id,
        'as_admin': as_admin
    }

    return execute_put(api_key, resource_url, body)


def delete_pulse_subscriber(api_key, pulse_id, user_id):

    resource_url = constants.PULSE_SUBSCRIBERS_BY_USER_ID.format(pulse_id, user_id)

    return execute_delete(api_key, resource_url)


def get_pulse_notes(api_key, pulse_id):

    resource_url = constants.PULSE_NOTES.format(pulse_id)

    return execute_get(api_key, resource_url)


def post_pulse_notes(api_key, 
        pulse_id, 
        title, 
        content, 
        owners_only = True, 
        user_id = None, 
        create_update = None):

    resource_url = constants.PULSE_NOTES.format(pulse_id)

    body = {
        'title': title,
        'content': content,
        'owners_only': owners_only
    }

    if user_id != None:
        body['user_id'] = user_id

        if create_update != None:
            body['create_update'] = create_update

    return execute_post(api_key, resource_url, body)


def put_pulse_note_by_id(
        api_key, 
        pulse_id,
        note_id,
        title,
        content,
        user_id = None,
        create_update = None):

    resource_url = constants.PULSE_NOTE_BY_ID.format(pulse_id, note_id)

    body = {
        'title': title,
        'content': content
    }

    if user_id != None:
        body['user_id'] = user_id

        if create_update != None:
            body['create_update'] = create_update

    return execute_put(api_key, resource_url, body)


def delete_pulse_note_by_id(api_key, pulse_id, note_id):

    resource_url = constants.PULSE_NOTE_BY_ID.format(pulse_id, note_id)

    return execute_delete(api_key, resource_url)


def get_pulse_updates(api_key, pulse_id, page = 1, limit = 25):

    resource_url = constants.PULSE_UPDATES.format(pulse_id)

    params = {
        'page': page,
        'limit': limit
    }

    return execute_get(api_key, resource_url, params)



import requests, json, time

from . import MondayApiError
from .graphql import *
from .constants import *

def execute_query(timeout: int = None, **kwargs):
    """Executes a graphql query via Rest.
    
        Parameters

            timeout : `int`
                The default timeout for Rest requests.
            kwargs : `dict`
                Optional keyword arguments

        Returns
        
            data : `dict`
                Response data in dictionary form.

        Optional Arguments

            api_key : `str`
                The monday.com API v2 user key.
            operation : `moncli.api_v2.graphql.GraphQLOperation`
                Perform request with input graphql operation.
            query_name: `str`:
                The name of the query to execute.
            operation_type: `moncli.api_v2.graphql.OperationType`:
                The type of graphql operation to perform (QUERY or MUTATION).
            fields: `list[str]`:
                List of fields to return.
            arguments: `dict`:
                Additional graphql arguments.
            query : `str`
                Perform request with raw graphql query string.
            variables : `dict`
                Variables added to the query.
    """

    api_key = kwargs.pop('api_key', None)
    if not api_key:
        from . import api_key

    if not timeout:
        from . import connection_timeout
        timeout = connection_timeout

    query_name = kwargs.pop('query_name', None)
    operation_type = kwargs.pop('operation_type', None)
    fields = kwargs.pop('fields', ())
    arguments = kwargs.pop('arguments', {})
    query = kwargs.pop('query', None)
    variables = kwargs.pop('variables', None)
    include_complexity = kwargs.pop('include_complexity', False)

    if not query:
        default_fields, default_arguments = QUERY_MAP.get(query_name, ([],{}))
        fields = get_field_list(default_fields, None, *fields)
        arguments = get_method_arguments(default_arguments, **arguments)
        query = GraphQLOperation(operation_type, query_name, FIELD_MAP, *fields, **arguments).format_body()

    if include_complexity:
        if 'mutation' in query:
            query = query.replace('query {', 'mutation { complexity { before, after }')
        else:
            query = query.replace('query {', 'query { complexity { before, after }')

    headers = { 'Authorization': api_key }
    data = { 'query': query, 'variables': variables }

    resp = requests.post(
        API_V2_ENDPOINT,
        headers=headers,
        data=data,
        timeout=timeout)

    return _process_repsonse(api_key, timeout, resp, data, **kwargs)[query_name]


def upload_file(file_path: str, timeout = 300, **kwargs):
    """Executes a graphql query to upload a file via Rest.
    
        Parameters

            timeout : `int`
                The default timeout for Rest requests.
            kwargs : `dict`
                Optional keyword arguments

        Returns

            data : `dict`
                Response data in dictionary form.

        Optional Arguments

            api_key : `str`
                The monday.com API v2 user key.
            query_name: `str`:
                The name of the query to execute.
            operation_type: `moncli.api_v2.graphql.OperationType`:
                The type of graphql operation to perform (QUERY or MUTATION).
            fields: `list[str]`:
                List of fields to return.
            arguments: `dict`:
                Additional graphql arguments.
    """

    api_key = kwargs.pop('api_key', None)
    if not api_key:
        from . import api_key

    query_name = kwargs.pop('query_name')
    fields = kwargs.pop('fields', None)
    default_fields, _ = QUERY_MAP.get(query_name, ([],{}))
    fields = get_field_list(default_fields, None, *fields)
    arguments = kwargs.pop('arguments', None)
    operation = GraphQLOperation(OperationType.MUTATION, query_name, FIELD_MAP, *fields, **arguments)
    operation.add_query_variable('file', 'File!')
    query = operation.format_body()
    
    headers = { 'Authorization': api_key }
    data = { 'query': query }
    files = { 'variables[file]': open(file_path, 'rb') }
    resp = requests.post(
        API_V2_FILE_ENDPOINT,
        headers=headers,
        data=data,
        files=files,
        timeout=timeout)

    return _process_repsonse(api_key, timeout, resp, data, **kwargs)[query_name]


def get_field_list(fields: list, prefix: str = None, *args):
    """Get list of query fields.

        Parameters
        
            fields : `list[str]`
                A predefined collection of default fields.
            prefix : `str`
                Any parent fields to be added to the field lookup.
            args : `tuple`
                Field names passed into the request.

        Returns
            
            fields : `list[str]`
                The fields to be retrieved for the query.
    """

    if isinstance(args, tuple):
        args = list(args)
    if not args:
        args = fields
    if 'id' not in args:
        args.append('id')
    if prefix:
        return ['{}.{}'.format(prefix, arg) for arg in args]
    return args


def get_method_arguments(mappings: dict, **kwargs):
    """Get mapped query field arguments.

        Parameters
        
            mappings : `dict`
                A predefined set of default mappings.
            kwargs : `dict`
                Argument parameters passed into the request.

        Returns
            
            arguments : `dict`
                Arguments to be used for the query fields.
    """
    
    optional_args = {}
    for key, value in mappings.items():   
        try:
            if isinstance(value, ArgumentValueKind):
                optional_args[key] = create_value(kwargs[key], value)
            elif isinstance(value, tuple):
                data = [create_value(item, value[1]).format() for item in kwargs[key]]
                optional_args[key] = create_value(data, value[0])
            elif isinstance(value, dict):
                optional_args[key] = get_method_arguments(value, **(kwargs[key]))
        except KeyError:
            # Ignore if no kwargs found
            continue
    for arg in optional_args:
        kwargs[arg] = optional_args[arg]
    return kwargs


def _process_repsonse(api_key: str, timeout: int, resp, data, **kwargs):
    """Process Rest graphql response/retry request."""

    text: dict = resp.json()

    if resp.status_code == 403 or resp.status_code == 500:
        raise MondayApiError(json.dumps(data), resp.status_code, '', [text['error_message']])
    if resp.status_code == 429:
        time.sleep(5)
        return execute_query(api_key, timeout, **kwargs)
    if text.__contains__('errors'):
        errors = text['errors']
        if not 'Query has complexity of' in errors[0]['message']: # May my sins be forgiven someday...
            error_query = json.dumps(data)
            status_code = resp.status_code
            errors = text['errors']
            raise MondayApiError(error_query, status_code, '', errors)
        # Wait for 5 seconds and try again in case of rate limiting... ^
        time.sleep(5)
        return execute_query(api_key, timeout, **kwargs)
    # Raise exception for parse errors.
    if text.__contains__('error_code'):
        error_query = json.dumps(data)
        raise MondayApiError(error_query, 400, text['error_code'], [text['error_message']])

    text: dict = resp.json()
    return text['data']




from schematics import types
from schematics.models import Model

from .. import api, entities as en


class _Update(Model):
    """Update base model"""
    id = types.StringType(required=True)
    creator_id = types.StringType(required=True)
    item_id = types.StringType(required=True)
    body = types.StringType()
    created_at = types.StringType()
    text_body = types.StringType()
    updated_at = types.StringType()


class Update(_Update):
    """An update
    
        Properties

            assets : `list[moncli.entities.Asset]`
                The update's assets/files.
            body: `str`
                The update's html formatted body.
            created_at: `str`
                The update's creation date.
            creator : `moncli.entities.User`
                The update's creator.
            creator_id : `str`
                The unique identifier of the update creator.
            id : `str`
                The update's unique identifier.
            item_id : `str`
                The update's item ID.
            replies : `list[moncli.entities.Reply]
                The update's replies.
            text_body : `str`
                The update's text body.
            updated_at : `str`
                The update's last edit date.
        
        Methods

            get_creator : `moncli.entities.User`
                Get the update's creator.
            add_reply : `moncli.entities.Reply`
                Add reply to update.
            get_replies : `list[moncli.entities.Reply]`
                Get update replies.
            add_file : `moncli.entities.Asset`
                Add a file to update.
            get_files : `list[moncli.entities.Asset]
                Get update's files.
            delete : `moncli.entity.Update`
                Delete an update.
    """

    def __init__(self, **kwargs):
        self.__creds = kwargs.pop('creds')
        self.__creator = None
        creator = kwargs.pop('creator', None)
        if creator:
            self.__creator = en.User(cred=self.__creds, **creator)
        self.__replies = None
        replies = kwargs.pop('replies', None)
        if replies:
            self.__replies = [Reply(creds=self.__creds, item_id=kwargs['item_id'], **reply) for reply in replies]
        self.__assets = None
        assets = kwargs.pop('assets', None)
        if assets:
            self.__assets = [en.Asset(creds=self.__creds, **asset) for asset in assets]
        super(Update, self).__init__(kwargs)

    def __repr__(self):
        o = self.to_primitive()
        if self.__replies and len(self.__replies) > 0:
            o['replies'] = self.__replies
        return str(o)

    @property
    def creator(self):
        """The update's creator."""
        if not self.__creator:
            self.__creator = self.get_creator()
        return self.__creator

    @property
    def replies(self):
        """The update's replies."""
        if not self.__replies:
            self.__replies = self.get_replies()
        return self.__replies

    @property
    def assets(self):
        """The update's assets/files."""
        return self.__assets

    def get_creator(self, *args):
        """Get the update's creator.
        
            Parameters

                args : `tuple`
                    The list of user fields to return.

            Returns

                user : `moncli.entities.User`
                    The update's creator.

            Return Fields

                account : `moncli.entities.Account`
                    The user's account.
                birthday : `str`
                    The user's birthday.
                country_code : `str`
                    The user's country code.
                created_at : `str`
                    The user's creation date.
                email : `str`
                    The user's email.
                enabled : `bool`
                    Is the user enabled or not.
                id : `str`
                    The user's unique identifier.
                is_admin: `bool`
                    Is the user a admin or not.
                is_guest : `bool`
                    Is the user a guest or not.
                is_pending : `bool`
                    Is the user a pending user.
                is_verified: `bool`
                    Is the user is verified
                is_view_only : `bool`
                    Is the user a view only user or not.
                join_date : `str`
                    The date the user joined the account.
                location : `str`
                    The user' location.
                mobile_phone : `str`
                    The user's mobile phone number.
                name : `str`
                    The user's name.
                phone : `str`
                    The user's phone number.
                photo_original : `str`
                    The user's photo in the original size.
                photo_small : `str`
                    The user's photo in small size (150x150).
                photo_thumb : `str`
                    The user's photo in thumbnail size (100x100).
                photo_thumb_small : `str`
                    The user's photo in small thumbnail size (50x50).
                photo_tiny : `str`
                    The user's photo in tiny size (30x30).
                teams : `list[moncli.entities.Team]`
                    The teams the user is a member in.
                time_zone_identifier : `str`
                    The user's time zone identifier.
                title : `str`
                    The user's title.
                url : `str`
                    The user's profile url.
                utc_hours_diff : `int`
                    The user's UTC hours difference.
        """

        user_data = api.get_users(
            *args,
            api_key=self.__creds.api_key_v2,
            ids=[int(self.creator_id)])[0]
        return en.User(creds=self.__creds, **user_data)


    def add_reply(self, body: str, *args):
        """Add reply to update.

            Parameters

                body : `str`
                    The text body of the reply.
                args : `tuple`
                    The list of reply fields to return.

            Returns

                update : `moncli.entities.Update`
                    The updated update.

            Return Fields

                assets : `list[moncli.entities.Asset]`
                    The update's assets/files.
                body: `str`
                    The update's html formatted body.
                created_at: `str`
                    The update's creation date.
                creator : `moncli.entities.User`
                    The update's creator
                creator_id : `str`
                    The unique identifier of the update creator.
                id : `str`
                    The update's unique identifier.
                item_id : `str`
                    The update's item ID.
                replies : `list[moncli.entities.Reply]
                    The update's replies.
                text_body : `str`
                    The update's text body.
                updated_at : `str`
                    The update's last edit date.
        """

        update_data = api.create_update(
            body,
            self.item_id,
            *args,
            api_key=self.__creds.api_key_v2,
            parent_id=self.id)
        return en.Update(creds=self.__creds, **update_data)


    def get_replies(self, *args):
        """Get update replies.
    
            Parameters

                args : `tuple`
                    The list of update fields to return.

            Returns

                replies : `list[moncli.entities.Reply]`
                    The update's replies.

            Return Fields

                body : `str`
                    The reply's html formatted body.
                created_at : `str`
                    The reply's creation date.
                creator : `moncli.entities.User`
                    The reply's creator.
                creator_id : `str`
                    The unique identifier of the reply creator.
                id : `str`
                    The reply's unique identifier.
                text_body : `str`
                    The reply's text body.
                updated_at : `str`
                    The reply's last edit date.
        """

         # Hard configure the pagination rate.
        page = 1
        page_limit = 500
        record_count = 500

        while record_count >= page_limit:
            updates_data = api.get_updates(
                'id', 'item_id', 
                *api.get_field_list(api.DEFAULT_REPLY_QUERY_FIELDS, 'replies', *args),
                api_key=self.__creds.api_key_v2,
                limit=page_limit,
                page=page,
                )
            
            try:
                target_update = [update for update in updates_data if update['id'] == self.id][0]
                return [Reply(creds=self.__creds, item_id=target_update['item_id'], **reply_data) for reply_data in target_update['replies']]
            except KeyError:
                if len(target_update) == 0:
                    page += 1
                    record_count = len(updates_data)
                    continue
        return [] 


    def add_file(self, file_path: str, *args):
        """Add a file to update.
    
            Parameters

                file_path : `str`
                    The path to the file to upload.
                args : `tuple`
                    The list of update fields to return.

            Returns

                asset : `moncli.entities.Asset`
                    The newly created asset.

            Return Fields

                created_at : `str`
                    The file's creation date.
                file_extension : `str`
                    The file's extension.
                file_size : `int`
                    The file's size in bytes.
                id : `str`
                    The file's unique identifier.
                name : `str`
                    The file's name.
                public_url : `str`
                    Public url to the asset, valid for 1 hour.
                uploaded_by : `moncli.entities.user.User`
                    The user who uploaded the file
                url : `str`
                    The user who uploaded the file
                url_thumbnail : `str`
                    Url to view the asset in thumbnail mode. Only available for images.  
        """

        asset_data = api.add_file_to_update(
            self.id,
            file_path,
            *args,
            api_key=self.__creds.api_key_v2)
        return en.Asset(**asset_data)


    def get_files(self, *args):
        """Get update's files.

            Parameters

                args : `tuple`
                    The list of asset fields to return.

            Returns

                assets : `list[moncli.entities.Asset]`
                    The update's files.

            Return Fields

                created_at : `str`
                    The file's creation date.
                file_extension : `str`
                    The file's extension.
                file_size : `int`
                    The file's size in bytes.
                id : `str`
                    The file's unique identifier.
                name : `str`
                    The file's name.
                public_url : `str`
                    Public url to the asset, valid for 1 hour.
                uploaded_by : `moncli.entities.user.User`
                    The user who uploaded the file
                url : `str`
                    The user who uploaded the file
                url_thumbnail : `str`
                    Url to view the asset in thumbnail mode. Only available for images.  
        """

         # Hard configure the pagination rate.
        page = 1
        page_limit = 500
        record_count = 500

        while record_count >= page_limit:
            updates_data = api.get_updates(
                'id', 
                *api.get_field_list(api.DEFAULT_ASSET_QUERY_FIELDS, 'assets', *args),
                api_key=self.__creds.api_key_v2, 
                limit=page_limit,
                page=page)
            
            try:
                target_update = [update for update in updates_data if update['id'] == self.id][0]
                return [en.Asset(creds=self.__creds, **asset_data) for asset_data in target_update['assets']]
            except KeyError:
                if len(target_update) == 0:
                    page += 1
                    record_count = len(updates_data)
                    continue
        return []   


    def delete(self, *args):
        """Delete the selected update.
    
            Parameters

                args : `tuple`
                    The list of update fields to return.
        
            Returns

                update : `moncli.entities.Update`
                    The deleted update.
        
            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The update's assets/files.
                body: `str`
                    The update's html formatted body.
                created_at: `str`
                    The update's creation date.
                creator : `moncli.entities.user.User`
                    The update's creator
                creator_id : `str`
                    The unique identifier of the update creator.
                id : `str`
                    The update's unique identifier.
                item_id : `str`
                    The update's item ID.
                replies : `list[moncli.entities.reply.Reply]
                    The update's replies.
                text_body : `str`
                    The update's text body.
                updated_at : `str`
                    The update's last edit date.
        """

        update_data = api.delete_update(
            self.id,
            *args,
            api_key=self.__creds.api_key_v2)
        return Update(creds=self.__creds, **update_data)


class _Reply(Model):
    """Reply base model"""
    id = types.StringType(required=True)
    creator_id = types.StringType(required=True)
    body = types.StringType()
    created_at = types.StringType()
    text_body = types.StringType()
    updated_at = types.StringType()


class Reply(_Reply):
    """A reply for an update

        Properties

            body : `str`
                The reply's html formatted body.
            created_at : `str`
                The reply's creation date.
            creator : `moncli.entities.User`
                The reply's creator.
            creator_id : `str`
                The unique identifier of the reply creator.
            id : `str`
                The reply's unique identifier.
            text_body : `str`
                The reply's text body.
            updated_at : `str`
                The reply's last edit date.

        Methods

            get_creator : `moncli.entities.User`
                Get the reply's creator.
    """

    def __init__(self, **kwargs):
        self.__creds = kwargs.pop('creds')
        self.__item_id = kwargs.pop('item_id')
        self.__creator = None
        super(Reply, self).__init__(kwargs)

    def __repr__(self):
        return str(self.to_primitive())

    @property
    def creator(self):
        """The reply's creator."""
        if not self.__creator:
            self.__creator = self.get_creator()
        return self.__creator
    
    def get_creator(self, *args):
        """Get the reply's creator.

            Parameters

                args : `tuple`
                    List of user return fields.

            Returns
            
                creator : `moncli.entities.User`
                    The creator of this update.

            Return Fields

                account : `moncli.entities.Account`
                    The user's account.
                birthday : `str`
                    The user's birthday.
                country_code : `str`
                    The user's country code.
                created_at : `str`
                    The user's creation date.
                email : `str`
                    The user's email.
                enabled : `bool`
                    Is the user enabled or not.
                id : `str`
                    The user's unique identifier.
                is_admin: `bool`
                    Is the user a admin or not.
                is_guest : `bool`
                    Is the user a guest or not.
                is_pending : `bool`
                    Is the user a pending user.
                is_view_only : `bool`
                    Is the user a view only user or not.
                join_date : `str`
                    The date the user joined the account.
                location : `str`
                    The user' location.
                mobile_phone : `str`
                    The user's mobile phone number.
                name : `str`
                    The user's name.
                phone : `str`
                    The user's phone number.
                photo_original : `str`
                    The user's photo in the original size.
                photo_small : `str`
                    The user's photo in small size (150x150).
                photo_thumb : `str`
                    The user's photo in thumbnail size (100x100).
                photo_thumb_small : `str`
                    The user's photo in small thumbnail size (50x50).
                photo_tiny : `str`
                    The user's photo in tiny size (30x30).
                teams : `list[moncli.entities.Team]`
                    The teams the user is a member in.
                time_zone_identifier : `str`
                    The user's time zone identifier.
                title : `str`
                    The user's title.
                url : `str`
                    The user's profile url.
                utc_hours_diff : `int`
                    The user's UTC hours difference.
        """

        user_data = api.get_users(
            *args,
            api_key=self.__creds.api_key_v2,
            ids=[int(self.creator_id)])[0]
        return en.User(creds=self.__creds, **user_data)



from .constants import BASE_URL, PORT, API_VERSION

def format_url(resource_url):

    return "{}:{}/{}/{}".format(
        BASE_URL, 
        PORT, 
        API_VERSION,
        resource_url)


def raise_mondayapi_error(method, resource_url, resp):
    
    raise MondayApiError(
        method, 
        format_url(resource_url), 
        resp.status_code,
         resp.text)


class MondayApiError(Exception):

    def __init__(self, method, error_url, error_code, message):

        self.method = method
        self.error_url = error_url
        self.error_code = error_code
        self.message = message


class MondayQueryParameters():

    def __init__(self, api_key):

        self.__dict = { 'api_key': api_key}


    def add_param(self, name, value):

        self.__dict[name] = value

    
    def add_params(self, params_dict):

        for key in params_dict:
            self.add_param(key, params_dict[key])

    
    def to_dict(self):

        return self.__dict


from schematics import types
from schematics.models import Model

from .. import api, entities as en

class _Asset(Model):
    """Asset base model."""
    id = types.StringType(required=True)
    created_at = types.DateType()
    file_extension = types.StringType()
    file_size = types.IntType()
    name = types.StringType()
    public_url = types.StringType()
    url = types.StringType()
    url_thumbnail = types.StringType()

    def __repr__(self):
        return str(self.to_primitive())


class Asset(_Asset):
    """A file uploaded to monday.com.
  
        Properties

            id : `str`
                The file's unique identifier.
            created_at : `str`
                The file's creation date.
            file_extension : `str`
                The file's extension.
            file_size : `int`
                The file's size in bytes.
            name : `str`
                The file's name.
            public_url : `str`
                Public url to the asset, valid for 1 hour.
            uploaded_by : `moncli.entities.User`
                The user who uploaded the file.
            url : `str`
                The file url.
            url_thumbnail : `str`
                Url to view the asset in thumbnail mode. Only available for images.

        Methods

            get_uploaded_by_user : `moncli.entities.User`
                Get the user who uploaded the file.
    """

    def __init__(self, **kwargs):
        self.__creds = kwargs.pop('creds', None)
        uploaded_by = kwargs.pop('uploaded_by', None)
        if uploaded_by:
            self.__uploaded_by = en.User(**uploaded_by)
        super(Asset, self).__init__(kwargs)

    @property
    def uploaded_by(self):
        """The user who uploaded the file."""

        if not self.__uploaded_by:
            self.__uploaded_by = self.get_uploaded_by_user()
        return self.__uploaded_by


    def get_uploaded_by_user(self, *args):
        """Get the user who uploaded the file.

            Parameters

                args : `tuple`
                    The list of user return fields.
    
            Returns

                user : `moncli.entities.User`
                    The user who uploaded the file.
            
            Return Fields

                account : `moncli.entities.Account`
                    The user's account.
                birthday : `str`
                    The user's birthday.
                country_code : `str`
                    The user's country code.
                created_at : `str`
                    The user's creation date.
                email : `str`
                    The user's email.
                enabled : `bool`
                    Is the user enabled or not.
                id : `str`
                    The user's unique identifier.
                is_guest : `bool`
                    Is the user a guest or not.
                is_pending : `bool`
                    Is the user a pending user.
                is_view_only : `bool`
                    Is the user a view only user or not.
                join_date : `str`
                    The date the user joined the account.
                location : `str`
                    The user' location.
                mobile_phone : `str`
                    The user's mobile phone number.
                name : `str`
                    The user's name.
                phone : `str`
                    The user's phone number.
                photo_original : `str`
                    The user's photo in the original size.
                photo_small : `str`
                    The user's photo in small size (150x150).
                photo_thumb : `str`
                    The user's photo in thumbnail size (100x100).
                photo_thumb_small : `str`
                    The user's photo in small thumbnail size (50x50).
                photo_tiny : `str`
                    The user's photo in tiny size (30x30).
                teams : `list[moncli.entities.Team]`
                    The teams the user is a member in.
                time_zone_identifier : `str`
                    The user's time zone identifier.
                title : `str`
                    The user's title.
                url : `str`
                    The user's profile url.
                utc_hours_diff : `int`
                    The user's UTC hours difference.
        """
        
        user_data = api.get_assets(
            *api.get_field_list(api.DEFAULT_USER_QUERY_FIELDS, 'uploaded_by', *args),
            api_key=self.__creds.api_key_v2,
            ids=[self.id])[0]['uploaded_by']
        return en.User(**user_data)


from .graphql import ArgumentValueKind

## API Configurations
API_V2_ENDPOINT = 'https://api.monday.com/v2'
API_V2_FILE_ENDPOINT = 'https://api.monday.com/v2/file'

## Operation methods
# Boards
BOARDS = 'boards'
CREATE_BOARD = 'create_board'
ARCHIVE_BOARD = 'archive_board'
ADD_SUBSCRIBERS_TO_BOARD = 'add_subscribers_to_board'
DELETE_SUBSCRIBERS_FROM_BOARD = 'delete_subscribers_from_board'
# Columns
CREATE_COLUMN = 'create_column'
CHANGE_COLUMN_TITLE = 'change_column_title'
CHANGE_COLUMN_VALUE = 'change_column_value'
CHANGE_SIMPLE_COLUMN_VALUE = 'change_simple_column_value'
CHANGE_MULTIPLE_COLUMN_VALUES = 'change_multiple_column_values'
# Groups
DUPLICATE_GROUP = 'duplicate_group'
CREATE_GROUP = 'create_group'
ARCHIVE_GROUP = 'archive_group'
DELETE_GROUP = 'delete_group'
# Items
ITEMS = 'items'
ITEMS_BY_COLUMN_VALUES = 'items_by_column_values'
ITEMS_BY_MULTIPLE_COLUMN_VALUES = 'items_by_multiple_column_values'
CREATE_ITEM = 'create_item'
CREATE_SUBITEM = 'create_subitem'
CLEAR_ITEM_UPDATES = 'clear_item_updates'
MOVE_ITEM_TO_GROUP = 'move_item_to_group'
ARCHIVE_ITEM = 'archive_item'
DELETE_ITEM = 'delete_item'
DUPLICATE_ITEM = 'duplicate_item'
# Updates
UPDATES = 'updates'
CREATE_UPDATE = 'create_update'
DELETE_UPDATE = 'delete_update'
# Notifications
CREATE_NOTIFICATION = 'create_notification'
# Tags
TAGS = 'tags'
CREATE_OR_GET_TAG = 'create_or_get_tag'
# Users 
USERS = 'users'
# Teams
TEAMS = 'teams'
ADD_TEAMS_TO_WORKSPACE = 'add_teams_to_workspace'
DELETE_TEAMS_FROM_WORKSPACE = 'delete_teams_from_workspace'
# Me
ME = 'me'
# Account
ACCOUNT = 'account'
# Webhooks
CREATE_WEBHOOK = 'create_webhook'
DELETE_WEBHOOK = 'delete_webhook'
# Workspace
CREATE_WORKSPACE = 'create_workspace'
ADD_USERS_TO_WORKSPACE = 'add_users_to_workspace'
DELETE_USERS_FROM_WORKSPACE = 'delete_users_from_workspace'

# Files
ASSETS = 'assets'
ADD_FILE_TO_UPDATE = 'add_file_to_update'
ADD_FILE_TO_COLUMN = 'add_file_to_column'


## Default Fields
DEFAULT_BOARD_QUERY_FIELDS = [
    'id', 
    'name', 
    'board_folder_id', 
    'board_kind',
    'communication', 
    'description', 
    'permissions',
    'pos',
    'state',
    'workspace_id',
    'updated_at'
]

DEFAULT_ACTIVITY_LOG_QUERY_FIELDS = [
    'account_id',
    'created_at',
    'data',
    'entity',
    'event',
    'id',
    'user_id'
]

DEFAULT_BOARD_VIEW_QUERY_FIELDS = [
    'id',
    'name',
    'settings_str',
    'type'
]

DEFAULT_COLUMN_QUERY_FIELDS = [
    'id',
    'title',
    'settings_str',
    'archived',
    'type',
    'width'
]

DEFAULT_ITEM_QUERY_FIELDS = [
    'id',
    'name',
    'created_at',
    'creator_id',
    'updated_at',
    'state'
]

DEFAULT_GROUP_QUERY_FIELDS = [
    'id', 
    'title', 
    'archived', 
    'color', 
    'deleted', 
    'position'
]

DEFAULT_COLUMN_VALUE_QUERY_FIELDS = [
    'id',
    'title',
    'text',
    'value',
    'additional_info'
]

DEFAULT_TAG_QUERY_FIELDS = [
    'id',
    'name',
    'color'
]

DEFAULT_NOTIFICATION_QUERY_FIELDS = [
    'id',
    'text'
]

DEFAULT_UPDATE_QUERY_FIELDS = [
    'id',
    'body',
    'created_at',
    'creator_id',
    'item_id',
    'text_body',
    'updated_at'
]

DEFAULT_REPLY_QUERY_FIELDS = [
    'id',
    'body',
    'created_at',
    'creator_id',
    'text_body',
    'updated_at'
]

DEFAULT_ASSET_QUERY_FIELDS = [
    'id',
    'name',
    'url',
    'file_size',
    'file_extension'
]

DEFAULT_WEBHOOK_QUERY_FIELDS = [
    'id',
    'board_id'
]

DEFAULT_WORKSPACE_QUERY_FIELDS = [
    'id',
    'name',
    'kind',
    'description'
]

DEFAULT_USER_QUERY_FIELDS = [
    'id',
    'name',
    'url',
    'email',
    'enabled',
    'birthday',
    'country_code',
    'is_admin',
    'is_guest',
    'is_pending',
    'is_verified',
    'is_view_only',
    'join_date',
    'location',
    'mobile_phone',
    'phone',
    'photo_original',
    'photo_thumb',
    'photo_tiny',
    'time_zone_identifier',
    'title',
    'utc_hours_diff'
]

DEFAULT_TEAM_QUERY_FIELDS = [
    'id',
    'name',
    'picture_url'
]

DEFAULT_ACCOUNT_QUERY_FIELDS = [
    'id',
    'name',
    'first_day_of_the_week',
    'logo',
    'show_timeline_weekends',
    'slug'
]

DEFAULT_PLAN_QUERY_FIELDS = [
    'max_users',
    'period',
    'tier',
    'version'
]


FIELD_MAP = {
    'account': DEFAULT_ACCOUNT_QUERY_FIELDS,
    'activity_logs': DEFAULT_ACTIVITY_LOG_QUERY_FIELDS,
    'assets': DEFAULT_ASSET_QUERY_FIELDS,
    'board': DEFAULT_BOARD_QUERY_FIELDS,
    'columns': DEFAULT_COLUMN_QUERY_FIELDS,
    'column_values': DEFAULT_COLUMN_VALUE_QUERY_FIELDS,
    'creator': DEFAULT_USER_QUERY_FIELDS,
    'group': DEFAULT_GROUP_QUERY_FIELDS,
    'groups': DEFAULT_GROUP_QUERY_FIELDS,
    'items': DEFAULT_ITEM_QUERY_FIELDS,
    'owner': DEFAULT_USER_QUERY_FIELDS,
    'replies': DEFAULT_REPLY_QUERY_FIELDS,
    'subscribers': DEFAULT_USER_QUERY_FIELDS,
    'subitems': DEFAULT_ITEM_QUERY_FIELDS,
    'tags': DEFAULT_TAG_QUERY_FIELDS,
    'teams': DEFAULT_TEAM_QUERY_FIELDS,
    'top_group': DEFAULT_GROUP_QUERY_FIELDS,
    'updates': DEFAULT_UPDATE_QUERY_FIELDS,
    'uploaded_by': DEFAULT_USER_QUERY_FIELDS,
    'users': DEFAULT_USER_QUERY_FIELDS,
    'views': DEFAULT_BOARD_VIEW_QUERY_FIELDS,
    'workspace': DEFAULT_WORKSPACE_QUERY_FIELDS,
}

## Field list and default arguments map
QUERY_MAP = {
    BOARDS: (
        DEFAULT_BOARD_QUERY_FIELDS, 
        {
            'limit': ArgumentValueKind.Int,
            'page': ArgumentValueKind.Int,
            'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
            'board_kind': ArgumentValueKind.Enum,
            'state': ArgumentValueKind.Enum,
            'activity_logs': {
                'limit': ArgumentValueKind.Int,
                'page': ArgumentValueKind.Int,
                'user_ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
                'column_ids': (ArgumentValueKind.List, ArgumentValueKind.String),
                'group_ids': (ArgumentValueKind.List, ArgumentValueKind.String),
                'item_ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
                'from': ArgumentValueKind.String,
                'to': ArgumentValueKind.String
            },
            'columns': {
                'ids': (ArgumentValueKind.List, ArgumentValueKind.String)
            },
            'groups': {
                'ids': (ArgumentValueKind.List, ArgumentValueKind.String),
                'items': {
                    'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
                    'limit': ArgumentValueKind.Int,
                    'page': ArgumentValueKind.Int
                }
            },
            'items': {
                'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
                'limit': ArgumentValueKind.Int,
                'page': ArgumentValueKind.Int
            },
            'order_by': ArgumentValueKind.Enum,
            'updates': {
                'limit': ArgumentValueKind.Int,
                'page': ArgumentValueKind.Int
            },
            'views': {
                'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
                'type': ArgumentValueKind.String
            }
        }),
    ADD_USERS_TO_WORKSPACE: (DEFAULT_USER_QUERY_FIELDS, {
        
    }),
    CREATE_BOARD: (
        DEFAULT_BOARD_QUERY_FIELDS, 
        {
            'workspace_id': ArgumentValueKind.Int,
            'template_id': ArgumentValueKind.Int
        }),
    ARCHIVE_BOARD: (DEFAULT_BOARD_QUERY_FIELDS, {}),
    ADD_SUBSCRIBERS_TO_BOARD: (
        DEFAULT_USER_QUERY_FIELDS, 
        {
            'kind': ArgumentValueKind.Enum
        }),
    DELETE_SUBSCRIBERS_FROM_BOARD: (DEFAULT_USER_QUERY_FIELDS, {}),
    CREATE_COLUMN: (
        DEFAULT_COLUMN_QUERY_FIELDS, 
        {
            'defaults': ArgumentValueKind.Json
        }),
    CHANGE_COLUMN_TITLE: (DEFAULT_COLUMN_QUERY_FIELDS, {}),
    CHANGE_COLUMN_VALUE: (DEFAULT_ITEM_QUERY_FIELDS, {}),
    CHANGE_SIMPLE_COLUMN_VALUE: (DEFAULT_ITEM_QUERY_FIELDS, {}),
    CHANGE_MULTIPLE_COLUMN_VALUES: (DEFAULT_ITEM_QUERY_FIELDS, {}),
    DUPLICATE_GROUP: (
        DEFAULT_GROUP_QUERY_FIELDS, 
        {
            'add_to_top': ArgumentValueKind.Bool,
            'group_title': ArgumentValueKind.String
        }),
    CREATE_GROUP: (DEFAULT_GROUP_QUERY_FIELDS, {}),
    ARCHIVE_GROUP: (DEFAULT_GROUP_QUERY_FIELDS, {}),
    DELETE_GROUP: (DEFAULT_GROUP_QUERY_FIELDS, {}),
    ITEMS: (
        DEFAULT_ITEM_QUERY_FIELDS, 
        {
            'assets': {
                'column_ids': (ArgumentValueKind.List, ArgumentValueKind.String)
            },
            'limit': ArgumentValueKind.Int,
            'page': ArgumentValueKind.Int,
            'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
            'newest_first': ArgumentValueKind.Bool,
            'column_values': {
                'ids': (ArgumentValueKind.List, ArgumentValueKind.Default),
            },
            'updates': {
                'limit': ArgumentValueKind.Int,
                'page': ArgumentValueKind.Int
            }
        }),
    ITEMS_BY_COLUMN_VALUES: (
        DEFAULT_ITEM_QUERY_FIELDS, 
        {
            'limit': ArgumentValueKind.Int,
            'page': ArgumentValueKind.Int,
            'column_type': ArgumentValueKind.String,
            'state': ArgumentValueKind.Enum,
            'column_values': {
                'ids': (ArgumentValueKind.List, ArgumentValueKind.Default),
            },
            'updates': {
                'limit': ArgumentValueKind.Int,
                'page': ArgumentValueKind.Int
            }
        }),
    ITEMS_BY_MULTIPLE_COLUMN_VALUES: (
        DEFAULT_ITEM_QUERY_FIELDS, 
        {
            'limit': ArgumentValueKind.Int,
            'page': ArgumentValueKind.Int,
            'column_type': ArgumentValueKind.String,
            'state': ArgumentValueKind.Enum
        }),
    CREATE_ITEM: (
        DEFAULT_ITEM_QUERY_FIELDS, 
        {
            'group_id': ArgumentValueKind.String,
            'column_values': ArgumentValueKind.Json
        }),
    CREATE_SUBITEM: (
        DEFAULT_ITEM_QUERY_FIELDS,
        {
            'column_values': ArgumentValueKind.Json
        }),
    CLEAR_ITEM_UPDATES: (DEFAULT_ITEM_QUERY_FIELDS, {}),
    MOVE_ITEM_TO_GROUP: (DEFAULT_ITEM_QUERY_FIELDS, {}),
    ARCHIVE_ITEM: (DEFAULT_ITEM_QUERY_FIELDS, {}),
    DELETE_ITEM: (DEFAULT_ITEM_QUERY_FIELDS, {}),
    DUPLICATE_ITEM: (
        DEFAULT_ITEM_QUERY_FIELDS, 
        {
            'board_id': ArgumentValueKind.Int,
            'with_updates': ArgumentValueKind.Bool,
            'item_id': ArgumentValueKind.Int
        }),
    UPDATES: (
        DEFAULT_UPDATE_QUERY_FIELDS, 
        {
            'limit': ArgumentValueKind.Int,
            'page': ArgumentValueKind.Int
        }),
    CREATE_UPDATE: (
        DEFAULT_UPDATE_QUERY_FIELDS, 
        {
            'parent_id': ArgumentValueKind.Int
        }),
    DELETE_UPDATE: (DEFAULT_UPDATE_QUERY_FIELDS, {}),
    CREATE_NOTIFICATION: (
        DEFAULT_NOTIFICATION_QUERY_FIELDS, 
        {
            'payload': ArgumentValueKind.Json
        }),
    TAGS: (
        DEFAULT_TAG_QUERY_FIELDS, 
        {
            'ids': (ArgumentValueKind.List, ArgumentValueKind.Int)
        }),
    CREATE_OR_GET_TAG: (
        DEFAULT_TAG_QUERY_FIELDS, 
        {
            'board_id': ArgumentValueKind.Int
        }),
    USERS: (
        DEFAULT_USER_QUERY_FIELDS,
        {
            'ids': (ArgumentValueKind.List, ArgumentValueKind.Int),
            'kind': ArgumentValueKind.Enum,
            'newest_first': ArgumentValueKind.String,
            'limit': ArgumentValueKind.Int
        }),
    TEAMS: (
        DEFAULT_TEAM_QUERY_FIELDS,
        {
            'ids': (ArgumentValueKind.List, ArgumentValueKind.Int)
        }),
    ADD_TEAMS_TO_WORKSPACE: (
        DEFAULT_TEAM_QUERY_FIELDS ,
         {

         }),
    DELETE_TEAMS_FROM_WORKSPACE: 
    (
        DEFAULT_TEAM_QUERY_FIELDS , 
        {
            
        }),
    ME: (DEFAULT_USER_QUERY_FIELDS, {}),
    ACCOUNT: (DEFAULT_ACCOUNT_QUERY_FIELDS, {}),
    CREATE_WEBHOOK: (
        DEFAULT_WEBHOOK_QUERY_FIELDS, 
        {
            'config': ArgumentValueKind.Json
        }),
    DELETE_WEBHOOK: (DEFAULT_WEBHOOK_QUERY_FIELDS, ),
    CREATE_WORKSPACE: (
        DEFAULT_WORKSPACE_QUERY_FIELDS, 
        {
            'description': ArgumentValueKind.String
        }),
    DELETE_USERS_FROM_WORKSPACE: (
        DEFAULT_USER_QUERY_FIELDS,
         {

         }),
    ASSETS: (DEFAULT_ASSET_QUERY_FIELDS, {}),
    ADD_FILE_TO_UPDATE: (DEFAULT_ASSET_QUERY_FIELDS, {}),
    ADD_FILE_TO_COLUMN: (DEFAULT_ASSET_QUERY_FIELDS, {})
}


class MondayApiError(Exception):
    """monday.com API exception

    __________
    Properties
    __________
    query : `str`
        The submitted query resulting in error.
    status_code : `int`
        The http status code of the response.
    error_code : `str`
        The monday.com API error code.
    messages : `list[str]`
        The list of error messages.
    """

    def __init__(self, query: str, status_code: int, error_code: str, messages: list):

        self.query = query
        self.status_code = status_code
        self.error_code = error_code
        self.messages = messages


class BaseCollection(object):
    
    def __init__(self, values: list = [], _type: type = str, key = lambda x: x):
        self._values = []
        self._key = key
        for value in values:
            if not isinstance(value, _type):
                raise TypeError('Expected data of type {}, got {} instead.'.format(_type.__name__, value.__class__.__name__))
            self._values.append(value)


    def __len__(self):
        return len(self._values)
        

    def __getitem__(self, index):
        return self._values[self._key(index)]


    def __setitem__(self, index, value):
        self._values[self._key(index)] = value


    def __iter__(self):
        for value in self._values:
            yield value

    
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.__len__() != len(other):
            return False
        for val in self._values:
            if val not in other:
                return False
        return True


    def __repr__(self):
        return str(self._values)

    
    def insert(self, index, value):
        self._values.insert(self._key(index), value)


    def append(self, value):
        self.insert(len(self._values), value)


import pytz, json, re
from datetime import datetime, timedelta, timezone
from enum import EnumMeta
from pytz.exceptions import UnknownTimeZoneError

from pycountry import countries
from schematics.exceptions import ConversionError, ValidationError
from schematics.types import BaseType

from . import enums

from . import entities as en
from .config import *


class MondayType(BaseType):

    native_type = None
    null_value = None
    allow_casts = ()
    native_default = None
    is_readonly = False

    def __init__(self, id: str = None, title: str = None, *args, **kwargs):
        self.original_value = None
        metadata = {}

        if not id and not title:
            raise TypeError('"id" or "title" parameter is required.')
        if id:
            metadata['id'] = id
        if title:
            metadata['title'] = title

        # Handle defaults
        default = kwargs.pop('default', None)
        if not default:
            default = self.native_default
        super(MondayType, self).__init__(*args, default=default, metadata=metadata, **kwargs)

    @property
    def changed_at(self):
        value = self.metadata.get('changed_at', None)
        if not value:
            return None
        changed_at = datetime.strptime(value, ZULU_FORMAT)
        utc = pytz.timezone('UTC')
        changed_at = utc.localize(changed_at, is_dst=False)
        return changed_at.astimezone(datetime.now().astimezone().tzinfo)

    def to_native(self, value, context=None):
        if not value:
            return self.native_default

        if not isinstance(value, en.cv.ColumnValue):
            if self.is_readonly:
                return value
            if isinstance(value, self.native_type):
                return self._process(value)
            if self.allow_casts and isinstance(value, self.allow_casts):
                return self._cast(value)
            raise ConversionError("Couldn't interpret '{}' as type {}.".format(str(value), self.native_type.__name__))

        self.metadata['id'] = value.id
        self.metadata['title'] = value.title
        settings = json.loads(value.settings_str) if value.settings_str else {}
        for k, v in settings.items():
            self.metadata[k] = v
        return self._process_column_value(value)

    def to_primitive(self, value, context=None):
        if self.null_value == None:
            return None
        if not value:
            return self.null_value
        # Assume that conversions are necessary before exporting
        value = self.to_native(value)
        return self._export(value)

    def _process(self, value):
        return value

    def _cast(self, value):
        return self.native_type(value)

    def _process_column_value(self, value: en.cv.ColumnValue):
        return value.value

    def _export(self, value):
        return value


class CheckboxType(MondayType):
    native_type = bool
    native_default = False
    allow_casts = (int, str)
    null_value = {}

    def _export(self, value):
        if value == True:
            return {'checked': 'true'}

class CountryType(MondayType):

    native_type = en.cv.Country
    allow_casts = (dict,)
    null_value = {}

    def _cast(self, value):
        try:
            return en.cv.Country(name = value['country'], code = value['code'])
        except KeyError:
            raise ConversionError('Unable to convert value "{}" to Country.'.format(value))
        
    def _export(self, value):
        if value.name and value.code:
            return {'countryName': value.name, 'countryCode': value.code}
        return self.null_value
    
    def validate_country(self, value): 
        country = countries.get(name=value.name)
        if not country:
            raise ValidationError('Value "{}" is not a valid country.'.format(value.name))
        code = countries.get(alpha_2=value.code)
        if not code:
            raise ValidationError('Value "{}" is not a valid alpha 2 country code.'.format(value.code))


class CreationLogType(MondayType):

    native_type = datetime
    native_default = None
    is_readonly = True

    
class DateType(MondayType):

    native_type = datetime
    allow_casts = (str, int)
    null_value = {}
    has_time = False

    def __init__(self, id: str = None, title: str = None, has_time: bool = False, *args, **kwargs):
        self.id = id
        self.title = title
        self.has_time = has_time
        super().__init__(self.id, self.title, *args, **kwargs)

    def _cast(self, value):
        if isinstance(value, int):
            try:
                return datetime.fromtimestamp(value)
            except ValueError:
                raise ConversionError('Invalid UNIX timestamp "{}".'.format(value))
        if isinstance(value, str):
            try:
                date = datetime.strptime(value.split(' ', 1)[0], DATE_FORMAT)
            except ValueError:
                raise ConversionError('Cannot convert value "{}" into Date.'.format(value))
            if self.has_time == True:
                try:
                    time = datetime.strptime(value.split(' ', 1)[1], TIME_FORMAT)
                    date = date + timedelta(hours=time.hour, minutes=time.minute, seconds=time.second)
                    return date
                except ValueError:
                    raise ConversionError('Cannot convert value "{}" into Time.'.format(value))
            return date

    def _export(self, value):
        if self.has_time == True:
            value = value.astimezone(timezone.utc)
            date = value.date().strftime(DATE_FORMAT)
            time = value.time().strftime(TIME_FORMAT)
            return {'date': date, 'time': time}
        return {'date': value.date().strftime(DATE_FORMAT), 'time': None}


class DependencyType(MondayType):

    native_type = list
    native_default = []
    null_value = {}

    def _process(self, value):
        value_list = []
        for data in value:
            try:
                value_list.append(int(data))
            except ValueError:
                raise ConversionError('Invalid item ID: "{}".'.format(data))
        return value_list
    
    def _export(self, value):
        return { 'item_ids': [data for data in value]}


class DropdownType(MondayType):
    native_type = list
    native_default = []
    null_value = {}
    element_type = str

    def __init__(self, id: str = None, title: str = None, as_enum: type = None, *args, **kwargs):
        if as_enum:
            self.element_type = as_enum
        super().__init__(id=id, title=title, *args, **kwargs)


    def _process(self, value):
        labels = {}
        return_list = []
        
        [labels.__setitem__(data['id'], data['name']) for data in self.metadata['labels']]
        for val in value:
            if isinstance(self.element_type, EnumMeta) and isinstance(val, self.element_type):
                return_list.append(val)
                continue
            try:
                label = labels[int(val)]
            except KeyError:
                raise ConversionError('Value "{}" is not a valid dropdown index.'.format(val))
            except ValueError:
                label = str(val)
                if not (label in labels.values()):
                    raise ConversionError('Value "{}" is not a valid dropdown label.'.format(val))
            if isinstance(self.element_type, EnumMeta):
                return_list.append(self.element_type(label))
            else:
                return_list.append(label)
        return return_list

    
    def _process_column_value(self, value: en.cv.ColumnValue):
        if isinstance(self.element_type,EnumMeta):
            return [self.element_type(data) for data in value.value]
        return super()._process_column_value(value)


    def _export(self, value):
        labels = {}
        for data in self.metadata['labels']:
            k,v = data.values()
            labels[v] = k
        
        labels_str = value
        if isinstance(self.element_type,EnumMeta):
            labels_str = [data.value for data in value]

        return_list =  [labels[label] for label in labels_str]
        return { 'ids': return_list}


class EmailType(MondayType):
    native_type = en.cv.Email
    null_value = {}
    allow_casts = (dict, )

    def _cast(self, value):
        try:
            return en.cv.Email(value['email'],value.get('text', value['email']))
        except KeyError:
            raise ConversionError('Cannot convert value "{}" to Email.'.format(value))

    def _export(self, value):
        return {'email': value.email, 'text': value.text}

    def validate_email(self, value):
        regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'
        if value.email and not re.fullmatch(regex, value.email):
            raise ValidationError('Value "{}" is not a valid email address.'.format(value))


class HourType(MondayType):
    native_type = en.cv.Hour
    null_value = {}
    allow_casts = (dict, )

    def _cast(self, value):
        try:
            return en.cv.Hour(value['hour'],value.get('minute', 0))
        except KeyError:
            raise ConversionError('Cannot convert value "{}" into Hour.'.format(value))
    def _export(self, value):
        return {'hour': value.hour, 'minute': value.minute}

    def validate_hour(self,value):
        if (value.hour > 23) or (value.hour < 0):
            raise ValidationError('Hour values must be between 0-23, not "{}".'.format(value.hour))
        if (value.minute > 59) or (value.minute < 0):
            raise ValidationError('Minute values must be between 0-59, not "{}".'.format(value.minute))


class ItemLinkType(MondayType):

    native_type = list
    native_default = []
    null_value = {}
    multiple_values = True
    element_type = int

    def __init__(self, id: str = None, title: str = None, multiple_values: bool = True, *args, **kwargs):
        self.multiple_values = multiple_values
        if not multiple_values:
            self.native_type = int
            self.native_default = None
            self.element_type = None
            self.allow_casts = (str,)
        super().__init__(id, title, *args, **kwargs)

    def _process(self, value):
        if not self.multiple_values:
            return value
        elif self.multiple_values:
            return_list = []   
            for data in value:
                try:
                    return_list.append(int(data))
                except ValueError:
                    raise ConversionError('Invalid item ID "{}".'.format(data))
            return return_list
            
    def _process_column_value(self, value: en.cv.ItemLinkValue):
        try:
            self.multiple_values = value.settings['allowMultipleItems']
        except KeyError:
            self.multiple_values = True
        if not self.multiple_values:
            self.native_type = int
            self.native_default = None
            self.element_type = None
            self.allow_casts = (str,)
            return value.value[0]
        return super()._process_column_value(value)

    def _export(self, value):
        if not self.multiple_values:
            return_value = int(value) if isinstance(value,int) else int(value[0])
            return {'item_ids': [return_value] }
        return {'item_ids': [int(data) for data in value ] }


class LastUpdatedType(MondayType):

    native_type = datetime
    native_default = None
    is_readonly = True


class LinkType(MondayType):
    
    native_type = en.cv.Link
    null_value = {}
    allow_casts = (dict,)

    def _cast(self, value):
        try:
            return en.cv.Link(value['url'],value['text'])
        except KeyError:
            raise ConversionError('Cannot convert value "{}" to Link.'.format(value))
    
    def _export(self, value):
        if value.url != None:
            return {'url': value.url, 'text': value.text}
        return self.null_value

    def validate_link(self,value):
        str_value = value.url
        if not (str_value.startswith('https://') or str_value.startswith('http://')):
            raise ValidationError('Value "{}" is not a valid URL link.'.format(value))
        return str_value


class LocationType(MondayType):

    native_type = en.cv.Location
    allow_casts = (dict,)
    null_value = {}

    def _cast(self, value):
        try:
            return en.cv.Location(lat=value['lat'],lng=value['lng'], address=value.get('address', None))
        except KeyError:
            raise ConversionError('Cannot convert "{}" to Location.'.format(value))

    def _export(self, value):
        if value.lat and value.lng:
            return { 'lat': value.lat,'lng': value.lng, 'address': value.address }
        return self.null_value
    
    def validate_location(self,value):
        if not (-90 <= value.lat <= 90):
            raise ValidationError('Value "{}" is not a valid Latitude.'.format(value))
        if not (-180 <= value.lng <= 180):
            raise ValidationError('Value "{}" is not a valid Longitude.'.format(value))


class LongTextType(MondayType):
    native_type = str
    allow_casts = (int, float)
    null_value = {}

    def _export(self, value):
        return {'text': value}


class NumberType(MondayType):
    native_type = (int, float)
    allow_casts = (str, )
    null_value = ""
    
    def _cast(self, value):
        try:
            number_value = int(value) if int(value) else float(value)
            return number_value
        except TypeError:
            raise ConversionError('Couldn\'t interpret str {} as int or float.'.format(value))
    
    def _export(self, value):
        return str(value)


class PeopleType(MondayType):

    native_type = list
    native_default = []
    null_value = {}
    element_type = en.cv.PersonOrTeam
    max_allowed = 0

    def __init__(self, id: str = None, title: str = None, max_allowed: int = 0, *args, **kwargs):
        self.max_allowed = max_allowed
        if max_allowed > 4 or max_allowed < 0:
            self.max_allowed = 0
        if max_allowed == 1:
            self.native_type = en.cv.PersonOrTeam
            self.native_default = None
            self.allow_casts = (dict, int, str)
            self.element_type = None
        super().__init__(id, title, *args, **kwargs)

    def _process(self, value):
        if isinstance(value,en.cv.PersonOrTeam):
            return value
        return_list = []
        for data in value:
            if isinstance(data,en.cv.PersonOrTeam):
                return_list.append(data)
                continue
            try:
                if isinstance(data, dict):
                    id = int(data['id'])
                    kind = enums.PeopleKind[data['kind']]
                    return_list.append(en.cv.PersonOrTeam(id=id,kind=kind))               
                elif isinstance(data, (str,int)):
                    return_list.append(en.cv.Person(int(data)))
                else:
                    raise ValueError('')
            except (ValueError, KeyError):
                raise ConversionError('Cannot convert value "{}" to Person or Team.'.format(data))
        return return_list


    def _cast(self, value):
        try:
            if isinstance(value, dict):
                return en.cv.PersonOrTeam(value['id'],enums.PeopleKind[value['kind']])
            if isinstance(value, (int,str)):
                return en.cv.Person(int(value))
        except (ValueError, KeyError):
            raise ConversionError('Cannot convert value "{}" to Person or Team.'.format(value))


    def _process_column_value(self, value: en.cv.ColumnValue):
        if 'max_people_allowed' in self.metadata:
            self.max_allowed = int(self.metadata['max_people_allowed'])
        if self.max_allowed == 1:
            self.native_type = en.cv.PersonOrTeam
            self.native_default = None
            self.allow_casts = (dict, int, str)
            self.element_type = None
            try:
                return value.value[0]
            except IndexError:
                return self.native_default
        return super()._process_column_value(value)

    def _export(self, value):
        if self.max_allowed == 1:
            return {'personsAndTeams': [{'id': value.id, 'kind': value.kind.name}]}
        personsAndTeams = [{'id': data.id, 'kind': data.kind.name} for data in value]
        return {'personsAndTeams': personsAndTeams}
        
    def validate_people(self, value):
        if self.max_allowed == 1 and not isinstance(value, en.cv.PersonOrTeam):
            raise ValidationError('Value contains too many Person or Team values: "{}".'.format(len(value)))
        if self.max_allowed in [2, 3] and len(value) > 4:
            raise ValidationError('Value contains too many Person or Team values: "{}".'.format(len(value)))


class PhoneType(MondayType):

    native_type = en.cv.Phone
    allow_casts = (dict,str)
    null_value = {}

    def _cast(self, value):
        if isinstance(value, dict):
            try:
                return en.cv.Phone(phone=value['phone'],code=value['code'])
            except KeyError:
                raise ConversionError('Unable to convert value "{}" to Phone.'.format(value))
        elif isinstance(value, str):
            values = value.split(" ",1)
            try:
                return en.cv.Phone(phone=values[0],code=values[1])
            except IndexError:
                raise ConversionError('Unable to convert value "{}" to Phone.'.format(value))

    def _export(self, value):
        if value.phone and value.code:
            return {'phone': value.phone, 'countryShortName': value.code}
        return self.null_value
    
    def validate_country_code(self, value):
        country = countries.get(alpha_2=value.code)
        if not country:
            raise ValidationError('Value "{}" is not a valid alpha 2 country code.'.format(value.code))


class RatingType(MondayType):
    native_type = int
    allow_casts = (str,)
    null_value = {}

    def _cast(self, value):
        try:
            return int(value)
        except ValueError:
            raise ConversionError('Value "{}" is not a valid rating.'.format(value))
    
    def _export(self, value):
        return { 'rating': value}


class StatusType(MondayType):
    native_type = str
    allow_casts = (int,)
    null_value = {}

    def __init__(self, id: str = None, title: str = None, as_enum: type = None ,  *args, **kwargs):
        if as_enum:
            if not isinstance(as_enum, EnumMeta):
                raise TypeError('Invalid type "{}" for status Enum.'.format(as_enum.__name__))
            self.choices = list(as_enum)
            for value in self.choices:
                if not isinstance(value.value, str):
                    raise TypeError('Invalid value "{}" for status Enum "{}".'.format(value.value, value.__class__.__name__))
            self.native_type =  as_enum
            self.allow_casts = (str,)
            
        super().__init__(id=id, title=title, *args, **kwargs)
    
    def _process(self, value):
        labels = self.metadata['labels']
        if self.native_type != str:
            return value
        try:
            int(value)
            try:
                return labels[value]
            except KeyError:
                raise ConversionError('Cannot find status label with index "{}".'.format(value))
        except ValueError:
            if value in labels.values():
                return value
            raise ConversionError('Cannot find status label "{}".'.format(value))
                
    def _cast(self, value):
        labels = self.metadata['labels']
        label = str(value)
        if self.native_type == str and isinstance(value,int):
            try:    
                return labels[label]
            except KeyError:
                raise ConversionError('Cannot find status label with index "{}".'.format(value))
            
        if isinstance(self.native_type, EnumMeta) and isinstance(value,str):
            try:
                return self.native_type(label)
            except ValueError:
                    raise ConversionError('Cannot find status label with index "{}".'.format(value))
    
    def _process_column_value(self, value: en.cv.ColumnValue):
        if self.native_type == str:
            self.choices = [value for value in self.metadata['labels'].values()]
        return super()._process_column_value(value)
                      
    def _export(self, value):
            labels = self.metadata['labels']
            if self.native_type == str:
                label = value
            elif isinstance(self.native_type, EnumMeta):
                label = value.value
            for index, value in labels.items():
                if value == label:
                    return {'index': int(index)}


class SubItemType(MondayType):

    native_type = list
    native_default = []
    null_value = {}
    is_readonly = True
 

class TextType(MondayType):
    native_type = str
    allow_casts = (int, float)
    null_value = ""


class TagsType(MondayType):
    native_type = list
    native_default = []
    null_value = {}

    def _process(self, value):
        for tag in value:
            try:
                return_value = [int(tag) for tag in value]
            except ValueError:
                raise ConversionError('Invalid Tag ID "{}".'.format(tag))
        return return_value
    
    def _export(self, value):
        return {'tag_ids': value}
        
        
class TimelineType(MondayType):

    native_type = en.cv.Timeline
    allow_casts = (dict,)
    null_value = {}

    def _cast(self, value):
        try:
            from_date = value['from']
            to_date = value['to']
            return en.cv.Timeline(from_date=from_date, to_date=to_date)
        except KeyError:
            raise ConversionError('Could not convert value "{}" to Timeline.'.format(value))

    def  _process_column_value(self, value: en.cv.ColumnValue):
        try:
            if value.settings['visualization_type']:
                self.metadata['is_milestone'] = True
        except KeyError:
            self.metadata['is_milestone'] = False
        return super()._process_column_value(value)

    def _export(self, value):
        if value.from_date == None or value.to_date == None:
            return {}
        return {'from': value.from_date.strftime(DATE_FORMAT), 'to': value.to_date.strftime(DATE_FORMAT)}
        
    def validate_timeline(self,value):
         if value.from_date > value.to_date:
            raise ValidationError('Timeline from date cannot be after to date.')


class TimeZoneType(MondayType):
    native_type = str
    null_value = {}

    def _export(self, value):
        return {'timezone': value}

    def validate_timezone(self, value):
        try:
            pytz.timezone(value)
        except (UnknownTimeZoneError):
            raise ValidationError('Unknown time zone "{}".'.format(value))


class WeekType(MondayType):
    
    native_type = en.cv.Week
    null_value = {}
    allow_casts = (dict,)

    def _cast(self, value):
        try:
            return en.cv.Week(value['start'],value['end'])
        except KeyError:
            raise ConversionError('Cannot convert value "{}" to Week.'.format(value))
    
    def _export(self, value):
        if not (value.start and value.end) :
            return self.null_value
        start =  value.start.strftime(DATE_FORMAT) 
        end  = value.end.strftime(DATE_FORMAT)
        return  {'week': {'startDate': start, 'endDate': end}}